# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module to handle the Simulink interface."""


from powertrain_build.lib import logger

LOGGER = logger.create_logger("simulink")


def get_interface(interface_data_types, dp_interface, hal_interface, sfw_interface, method_interface):
    """ Get interface combined for dp, hal and sfw

    Args:
        interface_data_types (dict): User defined interface data types
        dp_interface (dict): DP interface
        hal_interface (dict): HAL interface
        sfw_interface (dict): SFW interface
        method_interface (dict): Method interface
    Returns:
        output ([dict]): Combined interface
    """
    output = []
    output = add_dp(output, interface_data_types, split_interface(dp_interface))
    output = add_api(output, interface_data_types, split_interface(hal_interface))
    output = add_api(output, interface_data_types, split_interface(sfw_interface))
    output = add_methods(output, interface_data_types, method_interface)

    # Cannot have a completely empty adapters file, CSP will break
    if not output:
        if dp_interface:
            output = ensure_raster(output, list(dp_interface.keys())[0] + '_Nothing__Always')
        elif hal_interface:
            output = ensure_raster(output, list(hal_interface.keys())[0] + '_Nothing_Always')
        elif sfw_interface:
            output = ensure_raster(output, list(sfw_interface.keys())[0] + '_Nothing_Always')
    return output


def split_interface(interface):
    """Takes a raster interface and splits it based on read strategies.

    Args:
        interface (dict): DP/HAL/SFW raster interface.
    Returns:
        strategy_split_interface (dict): DP/HAL/SFW interface divided by read strategy.
    """
    strategy_split_interface = {}
    for raster, raster_data in interface.items():
        for port_type, signals in raster_data.items():
            for signal_spec in signals:
                interface_name = signal_spec.get('api') or signal_spec.get('domain').replace('_', '')
                new_raster = '_'.join([raster, interface_name, signal_spec['strategy']])
                if new_raster not in strategy_split_interface:
                    strategy_split_interface[new_raster] = {p: [] for p in raster_data}
                strategy_split_interface[new_raster][port_type].append(signal_spec)
    return strategy_split_interface


def ensure_raster(output, raster):
    """ Ensure raster exists in the output

    Args:
        output ([dict]): Combined interface
        raster (str): Name of raster
    Returns:
        output ([dict]): Combined interface
    """
    for adapter in output:
        if adapter['name'] == raster:
            return output
    output.append(
        {
            "name": raster,
            "ports": {
                "in": {},
                "out": {}
            }
        }
    )
    return output


def get_adapter(interface, raster):
    """ Get adapter

    Args:
        interface (dict): Combined interface
        raster (str): Name of raster
    Returns:
        adapter (dict): Adapter for the raster already in the interface
    """
    for adapter in interface:
        if adapter['name'] == raster:
            return adapter
    raise KeyError(raster)


def add_dp(output, interface_data_types, interface):
    """ Adds the DP interface to the combined interface

    Args:
        output ([dict]): Combined interface
        interface_data_types (dict): User defined interface data types
        interface (dict): DP interface
    Returns
        output (dict): Combined interface
    """
    ports = {
        "consumer": "in",
        "producer": "out"
    }
    for raster, raster_data in interface.items():
        if not isinstance(raster_data, dict):
            LOGGER.debug('Ignoring metadata: %s', raster_data)
            continue
        output = ensure_raster(output, raster)
        adapter = get_adapter(output, raster)
        for port_type, signals in raster_data.items():
            for signal in signals:
                data_type = signal['variable_type']
                primitive = ['signals']
                primitive.append(signal['domain'])
                if signal['group'] is not None:
                    primitive.append(signal['group'])
                primitive.append(signal['property'])
                adapter['ports'][ports[port_type]][signal['variable']] = {
                    'primitive': '.'.join(primitive),
                    'type': data_type
                }
                if 'enums' in interface_data_types and data_type in interface_data_types['enums']:
                    csp_enum_definition = add_csp_enum_def(data_type, interface_data_types['enums'][data_type])
                    if 'types' not in adapter:
                        adapter['types'] = []
                    if csp_enum_definition not in adapter['types']:
                        adapter['types'].append(csp_enum_definition)
    return output


def add_api(output, interface_data_types, interface):
    """ Adds the HAL/GenericApi/SFW interface to the combined interface

    Args:
        output ([dict]): Combined interface
        interface_data_types (dict): User defined interface data types
        interface (dict): HAL interface
    Returns
        output (dict): Combined interface
    """
    ports = {
        "consumer": "in",
        "producer": "out"
    }
    for raster, raster_data in interface.items():
        if not isinstance(raster_data, dict):
            LOGGER.debug('Ignoring metadata: %s', raster_data)
            continue
        output = ensure_raster(output, raster)
        adapter = get_adapter(output, raster)
        for port_type, signals in raster_data.items():
            for signal in signals:
                data_type = signal['variable_type']
                primitive = [signal['api'].lower()]
                primitive.append(signal['endpoint'].lower())
                if signal['property'] is not None:
                    primitive.append(signal['property'].lower())
                adapter['ports'][ports[port_type]][signal['variable']] = {
                    'primitive': '.'.join(primitive),
                    'type': data_type
                }
                if 'enums' in interface_data_types and data_type in interface_data_types['enums']:
                    csp_enum_definition = add_csp_enum_def(data_type, interface_data_types['enums'][data_type])
                    if 'types' not in adapter:
                        adapter['types'] = []
                    if csp_enum_definition not in adapter['types']:
                        adapter['types'].append(csp_enum_definition)
    return output


def add_methods(output, interface_data_types, methods):
    """ Adds the CSP method call interfaces to the combined interface.

    Args:
        output ([dict]): Combined interface
        interface_data_types (dict): Dict with enum definitions
        methods (dict): Methods used by the application
    Returns:
        output (dict): Combined interface
    """

    if methods == {}:
        return output
    output = ensure_raster(output, 'csp_methods')
    adapter = get_adapter(output, 'csp_methods')
    if 'methods' not in adapter:
        adapter['methods'] = []
    for method_data in methods.values():
        method_spec = {}
        method_spec['name'] = method_data['name']
        method_spec['primitive_method'] = method_data['primitive']
        method_spec['namespace'] = method_data['namespace']
        if 'description' in method_data:
            method_spec['description'] = method_data['description']
        for direction in ['in', 'out']:
            parameters = method_data['ports'][direction]
            if len(parameters) > 0:
                method_spec[direction] = []
            for param_name, param_data in parameters.items():
                param_spec = {}
                param_spec['name'] = param_name
                param_spec['primitive_name'] = param_data['primitive']
                data_type = param_data['type']
                param_spec['type'] = data_type
                method_spec[direction].append(param_spec)
                if data_type in interface_data_types.get('enums', {}):
                    csp_enum_definition = add_csp_enum_def(data_type, interface_data_types['enums'][data_type])
                    if 'types' not in adapter:
                        adapter['types'] = []
                    if csp_enum_definition not in adapter['types']:
                        adapter['types'].append(csp_enum_definition)
        adapter['methods'].append(method_spec)
    return output


def add_csp_enum_def(enum_name, interface_enum_definition):
    """ Returns a CSP style enumeration definition, given an interface style enum definition.

    Args:
        enum_name (str): Name of the enumeration
        interface_enum_definition (dict): Enumeration from interface data types
    Returns:
        csp_enum_definition (dict): Combined interface
    """
    default = {}
    enumerators = []
    for enum_member_definition in interface_enum_definition:
        if 'default' in enum_member_definition:
            default = {enum_member_definition['in']: enum_member_definition['out']}
        else:
            enumerators.append({enum_member_definition['in']: enum_member_definition['out']})

    csp_enum_definition = {
        'name': enum_name,
        'kind': 'enum',
        'enumerators': enumerators,
        'default': default
    }
    return csp_enum_definition



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for labelsplit files."""
import glob
import re
import json
import sys
from xml.etree import ElementTree
from pathlib import Path
from powertrain_build.feature_configs import FeatureConfigs
from powertrain_build.unit_configs import UnitConfigs
from powertrain_build.signal_interfaces import CsvSignalInterfaces
from powertrain_build.lib import helper_functions, logger

LOGGER = logger.create_logger(__file__)


class LabelSplit:
    """ Provides common LabelSplit functions for multiple repos.
    """
    def __init__(self, project, build_cfg, cfg_json, cmt_source_folder):
        """Read project configuration file to internal an representation.

        Args:
            project (str): Project name.
            build_cfg(BuildProjConfig): configures which units are active in the current project and where
                                        the code switch files are located.
            cfg_json(Path): Path to label split configuration file.
            cmt_source_folder (Path): Path to CMT source folder.
        """
        super().__init__()
        self.project = project
        self.build_cfg = build_cfg
        project_a2l_file_path = Path(self.build_cfg.get_src_code_dst_dir(),
                                     self.build_cfg.get_a2l_name())
        self.feature_cfg = FeatureConfigs(self.build_cfg)
        self.unit_cfg = UnitConfigs(self.build_cfg, self.feature_cfg)
        self.csv_if = CsvSignalInterfaces(self.build_cfg, self.unit_cfg)
        self.project_a2l_symbols = self.get_project_a2l_symbols(project_a2l_file_path)
        self.labelsplit_cfg = self.read_json(cfg_json)
        self.cmt_source_folder = cmt_source_folder

    @staticmethod
    def read_json(cfg_json):
        """Read label split configuration file from given location
           If the file does not exsit in the given location, program
           exits with error message.

        Args:
            cfg_json(Path): Path to label split configuration file
        Returns:
            labelsplit_cfg (dict): Dict of given file content
        """
        labelsplit_cfg = None
        if cfg_json.exists():
            with cfg_json.open() as json_file:
                labelsplit_cfg = json.load(json_file)
            return labelsplit_cfg
        LOGGER.error('Cannot find label split config file: %s', cfg_json)
        sys.exit(1)

    @staticmethod
    def get_project_a2l_symbols(project_a2l_file_path):
        """Get a list of calibration symbols found in a given project A2L file.

        Args:
            project_a2l_file_path (Path): Path to project A2L file.
        Returns:
            symbols_in_a2l (list): List of calibration symbols found in the project A2L file.
        """
        symbols_in_a2l = []

        with project_a2l_file_path.open() as a2l_fh:
            a2l_text = a2l_fh.read()

        calibration_blocks = re.findall(r'(?:\s*\n)*(\s*/begin (CHARACTERISTIC|AXIS_PTS)[\n\s]*'
                                        r'(\w+)([\[\d+\]]*).*?\n.*?/end \2)',
                                        a2l_text,
                                        flags=re.M | re.DOTALL)

        for blk in calibration_blocks:
            symbols_in_a2l.append(blk[2])

        return symbols_in_a2l

    @staticmethod
    def get_sgp_symbols(sgp_file: Path):
        """Get symbols and symbol_groups found in a given _sgp.xml file.

        Example output: {sVcExample: [(3, EC_EX_1), (4, EC_EX_2)]}, where the indices are column indices:
        1 -> symbol name (therefore not in list of symbol groups).
        2 -> diesel group.
        3 -> petrol group.
        4 -> hybrid group.
        5 -> subsystem (therefore not in list of symbol groups).

        Args:
            sgp_file (Path): Path to an _sgp.xml file.
        Returns:
            found_sgp_symbols (dict): A symbol to symbol_groups dictionary found in the sgp_file.
        """
        tree = ElementTree.parse(sgp_file)
        root = tree.getroot()
        search_string = '{{urn:schemas-microsoft-com:office:spreadsheet}}{tag}'
        label_sheet = root.find(search_string.format(tag='Worksheet'))
        table = label_sheet.find(search_string.format(tag='Table'))
        rows = table.findall(search_string.format(tag='Row'))

        found_sgp_symbols = {}
        for row in rows:
            symbol = None
            column_counter = 1
            cells = row.findall(search_string.format(tag='Cell'))
            for cell in cells:
                data = cell.find(search_string.format(tag='Data'))
                if data is not None:
                    # Sometimes there are spaces in the symbol cell
                    # Sometimes there is a weird \ufeff character (VcDebug_sgp.xml) in the symbol cell
                    value = data.text.replace(' ', '').replace('\ufeff', '')
                    if symbol is None:
                        symbol = value
                        found_sgp_symbols[symbol] = []
                    else:
                        new_index = search_string.format(tag='Index')
                        if new_index in cell.attrib:
                            column_counter = int(cell.attrib[new_index])
                        found_sgp_symbols[symbol].append((column_counter, value))
                column_counter += 1

        return found_sgp_symbols

    def get_sgp_symbol_group(self, symbol_groups_by_index):
        """Match _sgp.xml file indices (symbol groups) with a given project.

        Args:
            symbol_groups_by_index (list(tuple)): List of (index, symbol_group) pairs.
        Returns:
            symbol_group (str): The symbol group corresponding to the given project.
        """
        symbol_group = ''
        symbol_dict = self.labelsplit_cfg.get("SGP_SYMBOL_GROUPS")
        symbol_list = [val for key, val in symbol_dict.items() if key in self.project]
        if len(symbol_list) >= 1:
            for index, group in symbol_groups_by_index:
                if index == symbol_list[0]:
                    symbol_group = group
        else:
            LOGGER.error('Cannot match symbol group type for project: %s', self.project)
        return symbol_group

    def get_interface_symbols_and_groups(self, interface_dict, in_symbol_sgp_dict, out_symbol_sgp_dict):
        """Get a list of (symbol, symbol_group) pairs found in given interface and sgp files.

        Args:
            interface_dict (dict): interface to symbol map, matching a certain IO type.
            in_symbol_sgp_dict (dict): An input symbol to symbol_groups dictionary found in an sgp_file,
                to be compared with interface_dict inputs.
            out_symbol_sgp_dict (dict): An output symbol to symbol_groups dictionary found in an sgp_file,
                to be compared with interface_dict outputs.
        Returns:
            symbols_and_groups (list(tuple)): List of (symbol, symbol_group) pairs in: interface and sgp file.
        """
        symbols_and_groups = []
        for interface, symbol_data in interface_dict.items():
            for symbol in symbol_data.keys():
                debug_name = re.sub(r'\w(\w+)', r'c\1_db', symbol)
                switch_name = re.sub(r'\w(\w+)', r'c\1_sw', symbol)
                if 'Input' in interface and debug_name in in_symbol_sgp_dict and switch_name in in_symbol_sgp_dict:
                    debug_symbol_group = self.get_sgp_symbol_group(in_symbol_sgp_dict[debug_name])
                    switch_symbol_group = self.get_sgp_symbol_group(in_symbol_sgp_dict[switch_name])
                    symbols_and_groups.extend([(debug_name, debug_symbol_group), (switch_name, switch_symbol_group)])
                elif 'Output' in interface and debug_name in out_symbol_sgp_dict and switch_name in out_symbol_sgp_dict:
                    debug_symbol_group = self.get_sgp_symbol_group(out_symbol_sgp_dict[debug_name])
                    switch_symbol_group = self.get_sgp_symbol_group(out_symbol_sgp_dict[switch_name])
                    symbols_and_groups.extend([(debug_name, debug_symbol_group), (switch_name, switch_symbol_group)])
        return symbols_and_groups

    def get_debug_symbols_and_groups(self):
        """Get a list of (symbol, symbol_group) pairs found in project interface and VcDebug*_sgp.xml files.

        Returns:
            debug_symbols_and_groups (list(tuple)): List of (symbol, symbol_group) pairs in:
                interface and VcDebug*_sgp.xmlfiles.
        """
        _unused, dep, _unused_two, debug = self.csv_if.get_io_config()
        sgp_file_dict = self.labelsplit_cfg.get("SGP_FILE")
        debug_sgp_file = Path(sgp_file_dict.get('cfg_folder'), sgp_file_dict.get('debug'))
        debug_output_sgp_file = Path(sgp_file_dict.get('cfg_folder'), sgp_file_dict.get('debug_output'))
        dep_sgp_file = Path(sgp_file_dict.get('cfg_folder'), sgp_file_dict.get('dep'))
        dep_output_sgp_file = Path(sgp_file_dict.get('cfg_folder'), sgp_file_dict.get('dep_output'))
        debug_sgp_symbols = self.get_sgp_symbols(debug_sgp_file)
        debug_output_sgp_symbols = self.get_sgp_symbols(debug_output_sgp_file)
        dep_sgp_symbols = self.get_sgp_symbols(dep_sgp_file)
        dep_output_sgp_symbols = self.get_sgp_symbols(dep_output_sgp_file)

        symbols_and_groups_tmp = []
        debug_tmp = self.get_interface_symbols_and_groups(debug,
                                                          debug_sgp_symbols,
                                                          debug_output_sgp_symbols)
        dep_tmp = self.get_interface_symbols_and_groups(dep,
                                                        dep_sgp_symbols,
                                                        dep_output_sgp_symbols)
        symbols_and_groups_tmp.extend(debug_tmp)
        symbols_and_groups_tmp.extend(dep_tmp)

        debug_symbols_and_groups = []
        for symbol, symbol_group in symbols_and_groups_tmp:
            if symbol_group == '':
                LOGGER.info('Debug symbol %s is missing symbol group and will be removed.', symbol)
            else:
                debug_symbols_and_groups.append((symbol, symbol_group))

        return debug_symbols_and_groups

    def check_unit_par_file(self, unit):
        """Check <unit>_par.m file for default sgp symbol group.

        Args:
            unit (str): Current unit/model name.
        Returns:
            has_sgp_default (Bool): True/False if unit is associated with default sgp value.
            default_symbol_group (str): Name of default symbol group.
        """
        has_sgp_default = False
        default_symbol_group = ''
        base_search_string = r'SgpDefault\.{unit}\.[A-Za-z]+\s*=\s*[\'\"]([A-Za-z_]+)[\'\"]'
        search_string = base_search_string.format(unit=unit)

        non_existent_par_file = Path('non_existent_par_file.m')
        found_par_files = glob.glob('Models/*/' + unit + '/' + unit + '_par.m')
        if len(found_par_files) > 1:
            LOGGER.warning('Found more than one _par.m file, using %s', found_par_files[0])
        par_file = Path(found_par_files[0]) if found_par_files else non_existent_par_file

        if self.labelsplit_cfg.get("special_unit_prefixes"):
            for special_prefix in self.labelsplit_cfg.get("special_unit_prefixes"):
                if unit.startswith(special_prefix) and not par_file.is_file():
                    # Some units require special handling.
                    if '__' in unit:
                        parent = unit.replace('__', 'Mdl__')
                    else:
                        parent = unit + 'Mdl'
                    found_par_files = glob.glob('Models/*/' + parent + '/' + parent + '_par.m')
                    par_file = Path(found_par_files[0]) if found_par_files else Path(non_existent_par_file)
                    # Default symbol group is based on c-file name
                    c_name = re.sub('(Mdl)?(__.*)?', '', unit)
                    search_string = base_search_string.format(unit=c_name)

        if par_file.is_file():
            with par_file.open(encoding="latin-1") as par_fh:
                par_text = par_fh.read()
                sgp_default_match = re.search(search_string, par_text)
                if sgp_default_match is not None:
                    has_sgp_default = True
                    default_symbol_group = sgp_default_match.group(1)
        else:
            LOGGER.info('Missing _par file for model: %s', unit)

        return has_sgp_default, default_symbol_group

    def get_unit_sgp_file(self, unit):
        """Get path to <unit>_sgp.xml file.

        Args:
            unit (str): Current unit/model name.
        Returns:
            sgp_file (Path): Path to <unit>_sgp.xml file.
        """
        non_existent_sgp_file = Path('non_existent_sgp_file.xml')
        found_sgp_files = glob.glob('Models/*/' + unit + '/' + unit + '_sgp.xml')
        if len(found_sgp_files) > 1:
            LOGGER.warning('Found more than one _sgp.xml file, using %s', found_sgp_files[0])
        sgp_file = Path(found_sgp_files[0]) if found_sgp_files else Path(non_existent_sgp_file)

        if self.labelsplit_cfg.get("special_unit_prefixes"):
            for special_prefix in self.labelsplit_cfg.get("special_unit_prefixes"):
                if unit.startswith(special_prefix) and not sgp_file.is_file():
                    # Some units require special handling.
                    if '__' in unit:
                        parent = unit.replace('__', 'Mdl__')
                    else:
                        parent = unit + 'Mdl'
                    found_sgp_files = glob.glob('Models/*/' + parent + '/' + parent + '_sgp.xml')
                    sgp_file = Path(found_sgp_files[0]) if found_sgp_files else Path(non_existent_sgp_file)

        return sgp_file

    def get_unit_symbols_and_groups(self, unit, calibration_symbols):
        """Get a list of (symbol, symbol_group) pairs found in A2L, <unit>_sgp/par and config_<unit>.json files.

        Args:
            unit (str): Current unit/model name.
            calibration_symbols (list): All calibration symbols for the unit (from config_<unit>.json).
        Returns:
            unit_symbols_and_groups (list(tuple)): List of (symbol, symbol_group) pairs in: A2L, _sgp/_par and
            config files.
        """
        unit_symbols_and_groups = []
        has_sgp_default, default_symbol_group = self.check_unit_par_file(unit)
        sgp_file = self.get_unit_sgp_file(unit)

        if sgp_file.is_file():
            found_sgp_symbols = self.get_sgp_symbols(sgp_file)
        else:
            found_sgp_symbols = {}
            LOGGER.info('Missing _sgp file for model: %s', unit)

        for symbol in calibration_symbols:
            if symbol not in self.project_a2l_symbols:
                LOGGER.info('Symbol %s not in project A2L file and will be removed.', symbol)
                continue

            if symbol in found_sgp_symbols:
                symbol_group = self.get_sgp_symbol_group(found_sgp_symbols[symbol])
                unit_symbols_and_groups.append((symbol, symbol_group))
            elif has_sgp_default:
                if symbol.endswith('_sw') or symbol.endswith('_db'):
                    LOGGER.info('Debug symbol %s not in sgp file and will be removed.', symbol)
                else:
                    unit_symbols_and_groups.append((symbol, default_symbol_group))
            else:
                LOGGER.info('Symbol %s missing in _sgp file and lack SgpDefault value.', symbol)

        return unit_symbols_and_groups

    def get_calibration_constants(self):
        """Get all calibration symbols for each unit in the project.

        Returns:
            calibration_symbols_per_unit (dict): A unit to symbol list dictionary.
        """
        security_variables = self.labelsplit_cfg.get("security_variables")
        u_conf_dict = self.unit_cfg.get_per_cfg_unit_cfg()

        safe_calibration_symbols = {}
        for symbol, symbol_data in u_conf_dict['calib_consts'].items():
            if symbol not in security_variables:
                safe_calibration_symbols.update({symbol: symbol_data})

        calibration_symbols_per_unit = {}
        for symbol, symbol_data in safe_calibration_symbols.items():
            for unit, unit_data in symbol_data.items():
                if 'CVC_CAL' in unit_data['class']:
                    if unit in calibration_symbols_per_unit:
                        calibration_symbols_per_unit[unit].append(symbol)
                    else:
                        calibration_symbols_per_unit[unit] = [symbol]

        return calibration_symbols_per_unit

    def get_symbols_and_groups(self):
        """Get a list of (symbol, symbol_group) pairs found in A2L, <unit>_sgp/par and config_<unit>.json files.

        Returns:
            exit_code (int): 0/1 based on successful collection of symbols and symbol groups.
            all_symbols_and_groups (dict): A symbol to symbol_group dictionary for all A2L, _sgp/_par and config files,
        """
        exit_code = 0
        calibration_symbols_per_unit = self.get_calibration_constants()
        debug_symbols = self.get_debug_symbols_and_groups()

        all_symbol_and_group_pairs = []
        if self.labelsplit_cfg.get("project_symbols"):
            special_project_dict = self.labelsplit_cfg.get("project_symbols")
            pair_list = [val for key, val in special_project_dict.items() if key in self.project]
            if len(pair_list) == 1:
                symbol_and_group_pairs_list = pair_list[0].items()
                all_symbol_and_group_pairs += symbol_and_group_pairs_list
            elif len(pair_list) > 1:
                LOGGER.error('Project %s has does not follow the name rule', self.project)
                return 1, {}

        all_symbol_and_group_pairs.extend(debug_symbols)

        for unit, symbols in calibration_symbols_per_unit.items():
            if self.labelsplit_cfg.get("special_units"):
                special_unit_dict = self.labelsplit_cfg.get("special_units")
                if unit in special_unit_dict.keys():
                    # Some units require special handling.
                    LOGGER.warning('Found %s, assuming %s is used.', unit, special_unit_dict.get(unit))
                    labels = self.get_unit_symbols_and_groups(special_unit_dict.get(unit), symbols)
                else:
                    labels = self.get_unit_symbols_and_groups(unit, symbols)
            else:
                labels = self.get_unit_symbols_and_groups(unit, symbols)
            all_symbol_and_group_pairs.extend(labels)

        symbol_to_group_dict = {}
        for symbol, symbol_group in all_symbol_and_group_pairs:
            if symbol in symbol_to_group_dict:
                if symbol_to_group_dict[symbol] != symbol_group:
                    LOGGER.error('Symbol %s multiply defined with different symbol groups.', symbol)
                    exit_code = 1
            else:
                symbol_to_group_dict[symbol] = symbol_group

        return exit_code, symbol_to_group_dict

    def generate_label_split_xml_file(self, symbols_and_groups):
        """Generate a label split file, given a directory plus labels and groups to add.

        Args:
            symbols_and_groups (dict): A symbol to symbol_group dictionary given a project.
        Returns:
            exit_code (int): 0/1 based on successful generation of Labelsplit.xls.
        """
        errors = []
        project_root_dir = self.build_cfg.get_root_dir()
        cmt_output_folder = helper_functions.create_dir(Path(project_root_dir, 'output', 'CMT'))
        start_file_name = Path(self.cmt_source_folder, 'template_labelsplit_sgp_start.xml_')
        row_count_file_name = Path(cmt_output_folder, 'labelsplit_rowcount.xml_')
        start_2_file_name = Path(self.cmt_source_folder, 'template_labelsplit_sgp_start_2.xml_')
        label_split_rows_filename = Path(cmt_output_folder, 'labelsplit_rows.xml_')
        end_file_name = Path(self.cmt_source_folder, 'template_labelsplit_sgp_end.xml_')
        files_to_merge = [start_file_name, row_count_file_name, start_2_file_name,
                          label_split_rows_filename, end_file_name]

        with row_count_file_name.open('w', encoding="utf-8") as rc_fh:
            rc_fh.write(f'{len(symbols_and_groups) + 1}')  # header + data

        with label_split_rows_filename.open('w', encoding="utf-8") as lsrf_fh:
            for symbol, symbol_group in symbols_and_groups.items():
                if symbol_group == '':
                    errors.append(f'Missing symbol group for symbol: {symbol}')
                elif symbol_group == 'VCC_SPM_DEBUG':
                    LOGGER.info('Ignoring undistributed debug symbol: %s', symbol)
                else:
                    lsrf_fh.write(
                        '   <Row ss:AutoFitHeight="0">\n'
                        f'    <Cell><Data ss:Type="String">{symbol}'
                        '</Data><NamedCell ss:Name="_FilterDatabase"/></Cell>\n'
                        f'    <Cell ss:Index="5"><Data ss:Type="String">{symbol_group}'
                        '</Data><NamedCell ss:Name="_FilterDatabase"/></Cell>\n'
                        '   </Row>\n'
                    )

        if errors:
            LOGGER.error('\n'.join(errors))
            return 1

        output_file_name = Path(project_root_dir, 'output', 'CMT', 'Labelsplit.xls')
        with output_file_name.open('w', encoding="utf-8") as output_fh:
            for file_name in files_to_merge:
                with file_name.open(encoding="utf-8") as input_fh:
                    content = input_fh.read()
                    output_fh.write(content)
        LOGGER.info('Delivery to: %s', str(output_file_name))
        return 0



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Python module used for handling zone controller specifications"""
from ruamel.yaml import YAML
from powertrain_build.lib import logger
from powertrain_build.interface.base import BaseApplication

LOGGER = logger.create_logger("base")


class BadYamlFormat(Exception):
    """Exception to raise when in/out signal is not defined."""
    def __init__(self, message):
        self.message = message


class ZCAL(BaseApplication):
    """Zone controller abstraction layer"""

    def __repr__(self):
        """String representation of ZCAL"""
        return (
            f"<ZCAL {self.name}"
            f" app_side insignals: {len(self.signal_names['other']['insignals'])}"
            f" app_side outsignals: {len(self.signal_names['other']['outsignals'])}>"
        )

    def __init__(self, base_application):
        """Create the interface object

        Currently, there is no verification that the signal is used.

        TODO: Check if the signal is a set or a get somewhere. Maybe in here.

        Args:
            base_application (BaseApplication): Primary object of an interface
                                                Usually a raster, but can be an application or a model too.
        """
        self.name = ""
        self.zc_translations = {}
        self.composition_spec = {}
        self.signal_names = {}
        self.e2e_sts_signals = set()
        self.update_bit_signals = set()
        self.clear_signal_names()
        self.interface_enums = set()
        self.base_application = base_application
        self.translations_files = []
        self.device_domain = base_application.get_domain_mapping()
        self.signal_primitives_list = set()

    @staticmethod
    def read_translation_files(translation_files, keys=None):
        """ Searches translation files (yaml) for given keys at top level. Raises an
        error if conflicting configurations are found among the files. Return a dict
        containing the aggregated data found in the translation files for the keys
        provided in the function call.

        Args:
            translation_files (list): List of paths to files to search.
            keys (list): List of keys to search among translation files for (all present keys by default).

        Returns:
            specs (dict): Dictionary containing provided keys with the aggregated data
            stored under those keys in translation files.
        """

        specs = {}
        for translation_file in translation_files:
            with open(translation_file, encoding="utf-8") as translation:
                yaml = YAML(typ='safe', pure=True)
                raw = yaml.load(translation)
                used_keys = raw.keys() if keys is None else keys
                for key in used_keys:
                    new_elements = raw.get(key, None)
                    if isinstance(new_elements, list):
                        specs[key] = specs.get(key, []) + new_elements
                    elif isinstance(new_elements, dict):
                        specs[key] = specs.get(key, {})
                        if {**specs[key], **new_elements} != {**new_elements, **specs[key]}:
                            LOGGER.error(
                                "Conflicting configuration found for key '%s' among translation files: %s",
                                key,
                                translation_files,
                            )
                        specs[key].update(new_elements)
        return specs

    def add_signals(self, signals, signal_type="insignal", *args):
        """Add signal names and properties

        Args:
            signals (list(Signals)): Signals to use
            signal_type (str): 'insignals' or 'outsignals'
        """
        for signal in signals:
            if signal.name not in self.zc_translations:
                continue
            LOGGER.debug("Adding signal: %s", signal)
            for translation in self.zc_translations[signal.name]:
                signal_property = translation["property"]
                port_name = translation["port_name"]
                struct_name = translation["struct_name"]
                self.check_signal_property(port_name, struct_name, signal_property, signal_type)
                self.signal_names["zc"][signal_type].add(signal_property)
                self.signal_names["other"][signal_type].add(signal.name)
        for e2e_sts_signal_name in self.e2e_sts_signals:
            if e2e_sts_signal_name not in self.signal_names["other"]["insignals"]:
                LOGGER.warning("E2E check signal %s not used in any model.", e2e_sts_signal_name)
            else:
                self.signal_names["other"][signal_type].add(e2e_sts_signal_name)
        for update_bit_signal_name in self.update_bit_signals:
            if update_bit_signal_name not in self.signal_names["other"]["insignals"]:
                LOGGER.warning("Update bit signal %s not used in any model.", update_bit_signal_name)
            else:
                self.signal_names["other"][signal_type].add(update_bit_signal_name)
        LOGGER.debug("Registered signal names: %s", self.signal_names)

    def check_signal_property(self, port_name, struct_name, property_name, signal_type):
        """Check if we have only one signal written for the same property.

        Args:
            port_name (str): port name
            struct_name (str): signal struct name
            property_name (str): signal property
            signal_type (str): insignal or outsignal
        """
        signal_primitive_spec = ".".join([port_name, struct_name, property_name, signal_type])
        if signal_primitive_spec in self.signal_primitives_list and signal_type == "outsignal":
            error_msg = (f"You can't write {property_name} in {struct_name} as"
                         f" {signal_type} since this primitive has been used.")
            raise Exception(error_msg)
        self.signal_primitives_list.add(signal_primitive_spec)

    def parse_definition(self, definition):
        """Parses all translation files and populates class interface data.

        Args:
            definition (list(Path)): Definition files
        """
        raw = self.read_translation_files(definition)
        self.composition_spec = {
            key: value for key, value in raw.items() if key not in ["ports"]
        }
        ports_info = {}
        for port_name, port in raw.get("ports", {}).items():
            ports_info[port_name] = {k: v for k, v in port.items() if k not in ["element"]}
            signal_struct = port.get("element", {})
            direction = port.get("direction", None)
            if signal_struct:
                for element_name, element_data in signal_struct.items():
                    self.populate_signal_translations(port_name, element_name, element_data)
                # Loop inside function to make sure direction is consistent
                ports_info[port_name].update(self.get_port_info(port_name, direction, signal_struct))
        self.composition_spec["ports"] = ports_info

    @staticmethod
    def get_port_info(port_name, direction, signal_struct):
        """Extract port information from signal elements in port. Raises exception
        if signal elements are not exclusively sent in one direction.

        Args:
            port_name (str): Name of the port.
            direction (str): Direction of the port.
            signal_struct (dict): Signal dict containing list of signal elements.
        Returns:
            port_info (dict): Dict containing information if any elements
            should have an update bit associated with them.
        """
        if direction is None:
            raise BadYamlFormat(f'Port {port_name} is missing required property "direction".')

        port_info = {}
        update_elements = []
        end_to_end_elements = {}
        for element_name, element_data in signal_struct.items():
            for element in element_data:
                if "insignal" in element:
                    temp_dir = "IN"
                elif "outsignal" in element:
                    temp_dir = "OUT"
                else:
                    raise BadYamlFormat(f"in-/out-signal for element in { element_name } is missing.")
                # CLIENT and SERVER direction support both in and out signals
                if direction not in ["CLIENT", "SERVER"] and direction != temp_dir:
                    raise BadYamlFormat(f"Signal { element_name } has both in and out elements.")
                if element.get("updateBit", False) and element_name not in update_elements:
                    update_elements.append(element_name)
                if element.get("e2eStatus", False) and element_name not in end_to_end_elements:
                    end_to_end_elements[element_name] = {}

        if update_elements:
            port_info.update({"enable_update": update_elements})
        if end_to_end_elements:
            port_info.update({"end_to_end": [{k: v} for k, v in end_to_end_elements.items()]})
        return port_info

    def populate_signal_translations(self, port_name, element_name, element_data):
        """Populate class translations data.

        Args:
            port_name (str): Port name.
            element_name (str): Name of a port element.
            element_data (dict): Dict with signal structs to/from a port of an element.
        """
        enumerations = self.base_application.enumerations

        for signal_definition in element_data:
            if "insignal" in signal_definition:
                signal_name = signal_definition["insignal"]
                base_signals = self.base_application.insignals
            elif "outsignal" in signal_definition:
                signal_name = signal_definition["outsignal"]
                base_signals = self.base_application.outsignals
            else:
                raise BadYamlFormat(f"{ signal_name } is neither insignal nor outsignal in interface yaml file.")
            base_properties = None
            for base_signal in base_signals:
                if signal_name == base_signal.name:
                    matching_base_signal = base_signal
                    base_properties = self.base_application.get_signal_properties(
                        matching_base_signal
                    )
            if base_properties is None:
                raise BadYamlFormat(
                    f"{ signal_name } properties is missing. "
                    "Make sure that the model is in the raster file and that it is active in the current project."
                )

            if base_properties["type"] in enumerations:
                if 'init' in signal_definition:
                    init_value = signal_definition["init"]
                else:
                    if enumerations[base_properties['type']]['default_value'] is not None:
                        init_value = enumerations[base_properties['type']]['default_value']
                    else:
                        LOGGER.warning('Initializing enumeration %s to "zero".', base_properties['type'])
                        init_value = [
                            k for k, v in enumerations[base_properties['type']]['members'].items() if v == 0
                        ][0]
            else:
                init_value = signal_definition.get("init", 0)

            update_bit = signal_definition.get("updateBit", False)
            e2e_status = signal_definition.get("e2eStatus", False)
            group = signal_definition.get("group", element_name)
            translation = {
                "range": {
                    "min": base_properties.get("min", "-"),
                    "max": base_properties.get("max", "-")
                },
                "offset": base_properties.get("offset", "-"),
                "factor": base_properties.get("factor", "-"),
                "property": signal_definition["property"],
                "init": init_value,
                "port_name": port_name,
                "struct_name": element_name,
                "variable_type": base_properties.get("type"),
                "description": base_properties.get("description"),
                "unit": base_properties.get("unit"),
                "width": base_properties.get("width"),
                "debug": base_properties.get("debug", False),
                "dependability": e2e_status,
                "update_bit": update_bit
            }

            if signal_name not in self.zc_translations:
                self.zc_translations[signal_name] = []
            self.zc_translations[signal_name].append(translation)
            if update_bit:
                update_bit_property = f"{element_name}UpdateBit"
                update_signal_name = f"yVc{group}_B_{update_bit_property}"
                if signal_name == update_signal_name:
                    error_msg = f"Don't put updateBit status signals ({update_signal_name}) in yaml interface files."
                    raise BadYamlFormat(error_msg)
                self.zc_translations[update_signal_name] = [
                    {
                        "property": update_bit_property,
                        "variable_type": "Bool",
                        "property_interface_type": "Bool",
                        "offset": "-",
                        "factor": "-",
                        "range": {
                            "min": "-",
                            "max": "-",
                        },
                        "init": 1,
                        "port_name": port_name,
                        "struct_name": element_name,
                        "description": f"Update bit signal for signal {element_name}.",
                        "unit": None,
                        "dependability": False,
                        "update_bit": True
                    }
                ]
                self.update_bit_signals.add(update_signal_name)

            if e2e_status:
                if "outsignal" in signal_definition:
                    error_msg = "E2e status not expected for outsignals"
                    raise BadYamlFormat(error_msg)

                e2e_sts_property = f"{element_name}E2eSts"
                e2e_sts_signal_name = f"sVc{group}_D_{e2e_sts_property}"

                if signal_name == e2e_sts_signal_name:
                    error_msg = f"Don't put E2E status signals ({e2e_sts_signal_name}) in yaml interface files."
                    raise BadYamlFormat(error_msg)
                if e2e_sts_signal_name not in self.zc_translations:
                    self.zc_translations[e2e_sts_signal_name] = [
                        {
                            "property": e2e_sts_property,
                            "variable_type": "UInt8",
                            "property_interface_type": "UInt8",
                            "offset": 0,
                            "factor": 1,
                            "range": {
                                "min": 0,
                                "max": 255
                            },
                            "init": 255,
                            "port_name": port_name,
                            "struct_name": element_name,
                            "description": f"E2E status code for E2E protected signal(s) {signal_name}.",
                            "unit": None,
                            "debug": False,
                            "dependability": True,
                            "update_bit": False
                        }
                    ]
                    self.e2e_sts_signals.add(e2e_sts_signal_name)
                else:
                    translation = self.zc_translations[e2e_sts_signal_name][0]
                    translation["description"] = translation["description"][:-1] + f", {signal_name}."

    def clear_signal_names(self):
        """Clear signal names

        Clears defined signal names (but not signal properties).
        """
        self.signal_names = {
            "zc": {"insignals": set(), "outsignals": set()},
            "other": {"insignals": set(), "outsignals": set()}
        }

    def to_dict(self):
        """Method to generate dict to be saved as yaml

        Returns:
            spec (dict): Signalling specification
        """
        spec = {"consumer": [], "producer": []}

        signal_roles = ["consumer", "producer"]
        signal_types = ["insignals", "outsignals"]
        for signal_role, signal_type in zip(signal_roles, signal_types):
            for signal_name in self.signal_names["other"][signal_type]:
                for signal_spec in self.zc_translations[signal_name]:
                    spec[signal_role].append({**signal_spec, "variable": signal_name})

        return spec



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module for handling user defined data types, such as enumerations."""

import re
import time
from ruamel.yaml import YAML
from copy import deepcopy
from pathlib import Path
from powertrain_build.build_proj_config import BuildProjConfig
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.unit_configs import UnitConfigs


class UserDefinedTypes(ProblemLogger):
    """A class for accessing the project’s user defined data types (see :doc:`user_defined_types`)."""

    FILE_PREFIXES = ['udt_']  # Add more as needed
    UNDERLYING_DATA_TYPES = {  # Add more as needed
        'target_link': {
            'u32': 'UInt32',
            'i32': 'Int32',
            'u16': 'UInt16',
            'i16': 'Int16',
            'u8': 'UInt8',
            'i8': 'Int8'
        },
        'embedded_coder': {
            'u32': 'uint32_T',
            'i32': 'int32_T',
            'u16': 'uint16_T',
            'i16': 'int16_T',
            'u8': 'uint8_T',
            'i8': 'int8_T'
        }
    }
    SIMULINK_DATA_TYPES = {  # Add more as needed
        'target_link': {
            'Simulink.IntEnumType': 'Int32',
            'uint32': 'UInt32',
            'int32': 'Int32',
            'uint16': 'UInt16',
            'int16': 'Int16',
            'uint8': 'UInt8',
            'int8': 'Int8'
        },
        'embedded_coder': {
            'Simulink.IntEnumType': 'int32_T',
            'uint32': 'uint32_T',
            'int32': 'int32_T',
            'uint16': 'uint16_T',
            'int16': 'int16_T',
            'uint8': 'uint8_T',
            'int8': 'int8_T'
        }
    }

    def __init__(self, build_prj_config, unit_configs):
        """Class Initialization.

        Args:
            build_prj_config (BuildProjConfig): Instance holding information of where to find units configs to parse.
            unit_configs (UnitConfigs): Unit definitions.
        """
        super().__init__()
        if not isinstance(build_prj_config, BuildProjConfig) or not isinstance(unit_configs, UnitConfigs):
            err = (
                'Input arguments should be an instance of:'
                f'BuildProjConfig, not {type(build_prj_config)}'
                f'AND/OR UnitConfigs, not {type(unit_configs)}'
            )
            raise TypeError(err)

        start_time = time.time()
        self.info('  Start parsing files with user defined data types')
        self._build_prj_cfg = build_prj_config
        self._unit_configs = unit_configs
        self.enums_per_unit = {}
        self.common_enums = self._read_enums_from_definitions()
        self.structs_per_unit = {}
        self._parse_all_user_defined_types()
        self.common_header_files = []
        # Must run last to be able to compare with TL/EC data types
        self.all_enums = self._get_enumerations()
        self._interface_data_types = self._parse_interface_data_types()
        self.info('  Finished parsing files with user defined data types (in %4.2f s)', time.time() - start_time)

    @staticmethod
    def convert_interface_enum_to_simulink(interface_enum, underlying_data_type=None):
        """ Given an interface enumeration, convert it to a simulink parsed version.

        Args:
            interface_enum ([dict]): Interface enumeration definition.
            underlying_data_type (str): Underlying data type, usually not available for interface enumerations.
        Returns
            simulink_enum (dict): Simulink enumeration definition.
        """
        members = {}
        default_value = None
        # Assuming interface enumerations start at 0 counting upwards
        for idx, enum_member in enumerate(interface_enum):
            if 'default' in enum_member:
                default_value = enum_member['in']
            else:
                members[enum_member['in']] = idx
        simulink_enum = {
            'underlying_data_type': underlying_data_type,
            'members': members,
            'default_value': default_value
        }
        return simulink_enum

    def _parse_interface_data_types(self):
        """ Get interface data types.
        Also, runs validation tests.

        Returns:
            interface_data_types (dict): Specification interface data types.
        """
        interface_data_types = {}
        interface_data_types_path = Path(self._build_prj_cfg.get_root_dir(), 'conf.local', 'interface_data_types.yml')

        if not interface_data_types_path.is_file():
            self.warning('Cannot extract interface data types. File not found: %s', str(interface_data_types_path))
            return interface_data_types

        self.info('  Parsing file: %s', str(interface_data_types_path))
        with interface_data_types_path.open(mode='r', encoding='utf-8') as fh:
            yaml = YAML(typ='safe', pure=True)
            interface_data_types_yaml = yaml.load(fh)
            interface_data_types_tmp = interface_data_types_yaml['types']
        # Add more as needed
        if 'enums' in interface_data_types_tmp:
            valid_interface_enumerations = self._validate_interface_enumerations(interface_data_types_tmp['enums'])
            interface_data_types['enums'] = valid_interface_enumerations
        return interface_data_types

    def _parse_all_user_defined_types(self):
        """Parse all files containing user defined data types."""
        src_dirs = self._build_prj_cfg.get_unit_src_dirs()
        for unit, src_dir in src_dirs.items():
            self._parse_unit_user_defined_types(unit, src_dir)

    def _parse_unit_user_defined_types(self, unit, unit_src_dir):
        """Parse unit defined types for a given unit.

        Files to parse are found using expected prefixes, see FILE_PREFIXES.
        Data types to parse can be added as needed.

        Args:
            unit (str): Current unit/model name.
            unit_src_dir (str): Unit sourcecode directory.
        """
        self.enums_per_unit[unit] = {}
        self.structs_per_unit[unit] = {}
        found_files = [Path(unit_src_dir, unit.split('__')[0] + '.h')]
        for file_prefix in self.FILE_PREFIXES:
            found_files.extend(Path(unit_src_dir).glob(file_prefix + '*.h'))
        for found_file in found_files:
            self.debug('  Parsing file: %s', str(found_file))
            self._parse_target_link_enum(unit, found_file)
            self._parse_target_link_struct(unit, found_file)
        self._validate_project_enumerations()
        self._validate_project_structs()

    def _parse_target_link_struct(self, unit, header_file: Path):
        """Parse structs generated by TargetLink, given a header file.

        Args:
            unit (str): Current unit/model name.
            header_file (Path): Header file to parse.
        """
        with header_file.open(mode='r', encoding='ISO-8859-1') as hf_fh:
            header_file_content = hf_fh.read()
        structs = re.findall(
            r'^struct ([A-Za-z0-9_]+) {\s*\n'
            r'((?:\s*\w+ \w+;\s*\n)*)'
            r'};',
            header_file_content,
            flags=re.M
        )

        for struct_name, struct_content in structs:
            struct_members = re.findall(r'\s*(\w+) (\w+);', struct_content, flags=re.M)
            struct_dict = {
                'members': {variable: data_type for data_type, variable in struct_members}
            }

            if struct_name in self.structs_per_unit[unit]:
                if self.structs_per_unit[unit][struct_name]['members'] == struct_dict['members']:
                    self.info('Struct %s is multiply defined in %s, although they are consistent.', struct_name, unit)
                else:
                    self.critical('Found inconsistent multiply defined struct: %s, units: %s.', struct_name, [unit])
            else:
                self.structs_per_unit[unit][struct_name] = struct_dict.copy()

    def _validate_project_structs(self):
        """Checks for inconsistencies in struct definitions for all units in the project."""
        structs = {}
        for unit, struct_names in self.structs_per_unit.items():
            for struct_name, struct_data in struct_names.items():
                if struct_name in structs:
                    structs[struct_name]['units'].append(unit)
                    if not structs[struct_name]['members'] == struct_data['members']:
                        self.critical(
                            'Found inconsistent multiply defined struct: %s, units: %s.',
                            struct_name,
                            structs[struct_name]['units']
                        )
                else:
                    structs[struct_name] = struct_data.copy()
                    structs[struct_name]['units'] = [unit]

    def _validate_interface_enumerations(self, interface_enumerations):
        """ Given a specification for enumerations

        Args:
            interface_enumerations (dict): Interface enumerations.
        Returns
            valid_interface_enumerations (dict): Specification, valid interface enumerations.
        """
        valid_interface_enumerations = {}
        for enum_name, enum_data in interface_enumerations.items():
            if enum_name in valid_interface_enumerations:
                self.critical('%s is multiply defined in interface enumeration definitions.', enum_name)
            elif enum_name not in self.all_enums:
                self.critical('%s is not defined in the project', enum_name)
            else:
                converted = self.convert_interface_enum_to_simulink(
                    enum_data,
                    self.all_enums[enum_name]['underlying_data_type']
                )
                is_consistent = self._compare_enum_definitions(
                    self.all_enums[enum_name]['units'],
                    enum_name,
                    converted,
                    self.all_enums[enum_name]
                )
                if is_consistent:
                    valid_interface_enumerations[enum_name] = enum_data
        return valid_interface_enumerations

    def _validate_project_enumerations(self):
        """Checks for inconsistencies in enum definitions for all units in the project."""
        enumerations = {}
        for unit, enum_names in self.enums_per_unit.items():
            for enum_name, enum_data in enum_names.items():
                if enum_name in enumerations:
                    enumerations[enum_name]['units'].append(unit)
                    self._compare_enum_definitions(
                        enumerations[enum_name]['units'],
                        enum_name,
                        enumerations[enum_name],
                        enum_data
                    )
                else:
                    enumerations[enum_name] = enum_data.copy()
                    enumerations[enum_name]['units'] = [unit]

    def _compare_enum_definitions(self, units, enum_name, enum_one, enum_two):
        """Compare two enumeration definitions.

        Args:
            units ([str]): List of units using given enumerations.
            enum_name (str): Name of enumeration.
            enum_one (dict): Enumeration one to compare.
            enum_two (dict): Enumeration two to compare.
        Returns:
            is_consistent_enums (bool): True/False depending on enum consistency.
        """
        is_consistent_enums = True
        error_message = f'Found inconsistent multiply defined enumeration: {enum_name}, units: {units}.\n'
        if enum_one['underlying_data_type'] != enum_two['underlying_data_type']:
            is_consistent_enums = False
            error_message += (
                'underlying_data_type differs: '
                f"{enum_one['underlying_data_type']} vs. {enum_two['underlying_data_type']}.\n"
            )

        enum_one_members = list(enum_one['members'].items())
        enum_two_members = list(enum_two['members'].items())
        if len(enum_one_members) != len(enum_two_members):
            is_consistent_enums = False
            error_message += (
                f'Number of enum members differs: {len(enum_one_members)} vs. {len(enum_two_members)}.\n'
            )
        else:
            for idx, enum_one_member in enumerate(enum_one_members):
                if enum_one_member[0] != enum_two_members[idx][0] or enum_one_member[1] != enum_two_members[idx][1]:
                    is_consistent_enums = False
                    error_message += (
                        f'Enum member definition differs: {enum_one_members[idx]} vs. {enum_two_members[idx]}.\n'
                    )

        if enum_one['default_value'] != enum_two['default_value']:
            is_consistent_enums = False
            error_message += (
                'default_value differs: '
                f"{enum_one['default_value']} vs. {enum_two['default_value']}.\n"
            )

        if not is_consistent_enums:
            self.critical(error_message)

        return is_consistent_enums

    def _parse_target_link_enum(self, unit, header_file: Path):
        """Parse enumerations generated by TargetLink, given a header file.

        Args:
            unit (str): Current unit/model name.
            header_file (Path): Header file to parse.
        """
        with header_file.open(mode='r', encoding='ISO-8859-1') as hf_fh:
            header_file_content = hf_fh.read()
        enums = re.findall(
            r'^typedef enum ([A-Za-z0-9]+)_tag {\s*\n'
            r'((?:\s*\w+ = [-+]?\d+,?\s*\n)*)'
            r'} \1;',
            header_file_content,
            flags=re.M
        )

        for enum_name, enum_content in enums:
            enum_members_tmp = re.findall(r'\s*(\w+) = ([-+]?\d+)', enum_content, flags=re.M)
            enum_members = [(k, int(v)) for k, v in enum_members_tmp]
            if enum_name in self.common_enums:
                underlying_data_type = self.common_enums[enum_name]['underlying_data_type']
            else:
                self.warning('Calculating underlying data type for: %s', str(enum_name))
                underlying_data_type = self._calculate_underlying_data_type(unit, enum_name, enum_members)
            enum_dict = {
                'underlying_data_type': underlying_data_type,
                'members': {},
                'default_value': self.get_default_enum_value(unit, enum_name)
            }
            for key, value in enum_members:
                enum_dict['members'][key] = value

            if enum_name in self.enums_per_unit[unit]:
                is_consistent_enums = self._compare_enum_definitions(
                    [unit],
                    enum_name,
                    self.enums_per_unit[unit][enum_name],
                    enum_dict
                )
                if is_consistent_enums:
                    self.info(
                        'Enumeration %s is multiply defined in %s, although they are consistent.',
                        enum_name,
                        unit
                    )
            else:
                self.enums_per_unit[unit][enum_name] = enum_dict.copy()

    def _calculate_underlying_data_type(self, unit, enum_name, enum_members):
        """Calculate best fitting data type given an enum definition.

        Args:
            unit (str): Current unit/model name.
            enum_name (str): Current enum name.
            enum_members (list(tuple)): List of enum members, name value pairs.
        Returns:
            underlying_data_type (str): Best fitting data type.
        """
        code_generator = self._unit_configs.get_unit_code_generator(unit)
        enum_member_values = [member_value for member_name, member_value in enum_members]
        min_value = min(enum_member_values)
        max_value = max(enum_member_values)

        # TODO Consider forcing signed like CS (or ARXML, from database?) does, seems to be int8 specifically
        underlying_data_type = None
        if min_value >= 0:
            if max_value <= 255:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['u8']
            elif max_value <= 65535:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['u16']
            elif max_value <= 4294967295:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['u32']
        elif min_value >= -128:
            if max_value <= 127:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['i8']
            elif max_value <= 32767:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['i16']
            elif max_value <= 2147483647:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['i32']
        elif min_value >= -32768:
            if max_value <= 32767:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['i16']
            elif max_value <= 2147483647:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['i32']
        elif min_value >= -2147483648:
            if max_value <= 2147483647:
                underlying_data_type = self.UNDERLYING_DATA_TYPES[code_generator]['i32']

        if underlying_data_type is None:
            self.critical(
                'Unhandled enum size, name: %s, min: %s, max: %s. Valid types are uint8/16 and int8/16',
                enum_name,
                min_value,
                max_value
            )

        return underlying_data_type

    def _read_enums_from_definitions(self):
        """ Reads all enums from .m files where they are defined. Requires projectInfo.enumDefDir in
        ProjectCfg.json file.
        Files defining simulink class enumerations are expected to follow a template.

        Returns:
            common_enums (dict): Dictionary containing all enums defined in the enum definition directory.
        """
        common_enums = {}
        if self._build_prj_cfg.get_enum_def_dir() is None:
            self.warning(
                'Cannot parse enumerations from .m files. Missing "enumDefDir" in project config.'
            )
            return {}

        enum_files = Path(self._build_prj_cfg.get_enum_def_dir()).rglob('*.m')
        for enum_file in enum_files:
            with enum_file.open(mode='r', encoding='ISO-8859-1') as fh:
                enum_file_content = fh.read()
            enum_name = enum_file.stem

            members_part = re.search(
                r'\s+enumeration\s*\n'
                r'(\s*%.*|\s*\w+\s*\(\d+\)\s*(%.*)?\n)+'
                r'\s+end',
                enum_file_content,
                flags=re.MULTILINE
            )
            if not members_part:
                self.warning('Cannot extract enumeration members from %s', str(enum_file))
                continue
            members = re.findall(r'(\w+)\s*\((\d+)\)', members_part.group(0))
            member_dict = {enum_name + '_' + member[0]: int(member[1]) for member in members}
            enum_members = [(member_name, member_value) for member_name, member_value in member_dict.items()]

            get_default_value = re.search(
                r'function ([A-Za-z0-9]+) = getDefaultValue\(\)[\s\n]*'
                r'\1 = [A-Za-z0-9]+\.([A-Za-z0-9_]+);[\s\n]*'
                r'end',
                enum_file_content,
                flags=re.MULTILINE
            )
            if get_default_value is None:
                self.warning('Cannot extract default enumeration values in: %s', str(enum_file))
                continue
            default_value = get_default_value.groups()[1]
            default_member = f'{enum_name}_{default_value}'

            calculate_underlying_data_type = re.search(
                r'function ([A-Za-z0-9]+) = calculateUnderlyingDataType\(\)[\s\n]*'
                r'\1 = (true|false);[\s\n]*'
                r'end',
                enum_file_content,
                flags=re.MULTILINE
            )
            get_storage_type = re.search(
                r'classdef[\s]+[A-Za-z0-9]+[\s]*<[\s]*([A-Za-z0-9\.]+)[\s\n]*',
                enum_file_content,
                flags=re.MULTILINE
            )
            if get_storage_type is None:
                self.warning('Cannot extract storage type in: %s', str(enum_file))
                if calculate_underlying_data_type is None or calculate_underlying_data_type.groups()[1] == 'false':
                    self.warning('Calculating underlying data type for: %s', str(enum_file))
                    underlying_data_type = self._calculate_underlying_data_type('target_link', enum_name, enum_members)
                else:
                    underlying_data_type = self._calculate_underlying_data_type('target_link', enum_name, enum_members)
            else:
                underlying_data_type = self.SIMULINK_DATA_TYPES['target_link'][get_storage_type.groups()[0]]

            common_enums[enum_name] = {
                "underlying_data_type": underlying_data_type,
                "default_value": default_member,
                "members": member_dict
            }

        return common_enums

    def _get_enumerations(self):
        """Get all enumeration defined in the project, together with unit usage.

        Information already provided in self._validate_project_enumerations during initialization.

        Returns:
            enumerations (dict): Enumerations defined in the projects, including unit usage.
        """
        enumerations = {}
        for unit, enum_names in self.enums_per_unit.items():
            for enum_name, enum_data in enum_names.items():
                if enum_name in enumerations:
                    enumerations[enum_name]['units'].append(unit)
                else:
                    enumerations[enum_name] = deepcopy(enum_data)
                    enumerations[enum_name]['units'] = [unit]
        if self._build_prj_cfg.get_code_generation_config("includeAllEnums"):
            for enum_name, enum_data in self.common_enums.items():
                if enum_name not in enumerations:
                    self.warning(
                        "Enumeration %s is not used in any unit. Included since 'includeAllEnums' is set in config.",
                        enum_name
                    )
                    enumerations[enum_name] = deepcopy(enum_data)
                    enumerations[enum_name]['units'] = []
        return enumerations

    def get_default_enum_value(self, unit, enum_name):
        """Get default value of given enumeration name by searching in the unit configuration.

        Args:
            unit (str): Current unit/model name.
            enum_name (str): Name of an enumeration.
        Returns:
            default_value_str (str): Default value (string) for the given enumeration name.
                None if value could not be extracted.
        """
        # The config file is generated in one go, stop when first occurrence is found, cannot differ
        per_unit_cfg = self._unit_configs.get_per_unit_cfg()
        if unit in per_unit_cfg:
            u_cfg = per_unit_cfg[unit]
        else:
            # Should not happen
            self.warning(
                'Cannot extract default enumeration value for %s in %s. Unit is missing in the unit configuration.',
                enum_name,
                unit
            )
            return None

        for signal_type in ['inports', 'outports', 'local_vars', 'calib_consts']:
            for signal_data in u_cfg[signal_type].values():
                if enum_name == signal_data['type'] and 'default' in signal_data:
                    return f"{enum_name.upper()}_{signal_data['default'].upper()}"

        self.warning(
            'Cannot extract default enumeration value for %s in %s. '
            'Either the enumeration or its default value is missing in the unit configuration file. ',
            enum_name,
            unit
        )
        common_default = self.common_enums.get(enum_name, {}).get('default_value', None)
        if common_default is not None:
            return common_default.upper()
        return None

    def get_enumerations(self):
        """Get all enumeration defined in the project, together with unit usage.

        Returns:
            self.all_enums (dict): Enumerations defined in the projects, including unit usage.
        """
        return self.all_enums

    def get_interface_data_types(self):
        """Returns all interface data types"""
        return deepcopy(self._interface_data_types)

    def get_structs(self):
        """Get all structs defined in the project, together with unit usage.

        Information already provided in self._validate_project_structs during initialization.

        Returns:
            structs (dict): Structs defined in the projects, including unit usage.
        """
        structs = {}
        for unit, struct_names in self.structs_per_unit.items():
            for struct_name, struct_data in struct_names.items():
                if struct_name in structs:
                    structs[struct_name]['units'].append(unit)
                else:
                    structs[struct_name] = deepcopy(struct_data)
                    structs[struct_name]['units'] = [unit]
        return structs

    def _get_header_file_header(self, guard: str):
        """Get header for common header files.

        Args:
            guard (str): Guard for common header files.
        Returns:
            header (str): Header for common header files.
        """
        return (
            f'#ifndef {guard}\n'
            f'#define {guard}\n'
            f'{self._unit_configs.base_types_headers}'
        )

    def _generate_struct_header_file(self, file_path: Path):
        """ Generates a header file declaring all TargetLink structs used in the project.

        Args:
            file_path (Path): Path to file to generate.
        """
        structs = self.get_structs()
        guard = f"{file_path.stem.upper()}_H"
        with file_path.open('w') as fh:
            fh.write(self._get_header_file_header(guard))
            fh.write('#include "VcEnumerations.h"\n')  # struct members may be of enum type
            fh.write('/* VCC Structs */\n')
            for struct_name, struct_data in structs.items():
                fh.write('\n')
                fh.write(f'struct {struct_name} {{\n')
                for variable, data_type in struct_data['members'].items():
                    fh.write(f"   {data_type} {variable};\n")
                fh.write('};\n')
            fh.write(f'#endif /* {guard} */\n')

    def _generate_enum_header_file(self, file_path: Path):
        """ Generates a header file declaring all TargetLink enumerations used in the project.

        Args:
            file_path (Path): Path to file to generate.
        """
        enumerations = self.get_enumerations()
        rte_interface_enums = self._build_prj_cfg.get_code_generation_config("mapToRteEnums")
        if not self._build_prj_cfg.get_enum_def_dir() and rte_interface_enums:
            self.warning(
                "No enum definition directory specified. Cannot generate all enums in decinition directory."
            )
            rte_interface_enums = False
        guard = f"{file_path.stem.upper()}_H"
        with file_path.open('w') as fh:
            fh.write(self._get_header_file_header(guard))
            if rte_interface_enums:
                composition_name = self._build_prj_cfg.get_composition_config("softwareComponentName")
                fh.write(f'#include "Rte_{ composition_name }_Type.h"\n')
            fh.write('/* VCC Enumerations */\n')
            for enum_name, enum_data in enumerations.items():
                members = enum_data['members']
                member_strs = ', \n'.join([f"   {member.upper()} = {value}" for member, value in members.items()])
                enum_definition = (
                    f'\n'
                    f'typedef enum {enum_name}_tag {{\n'
                    f'{member_strs} \n'
                    f'}} {enum_name};\n'
                )
                if rte_interface_enums:
                    enum = self.common_enums[enum_name]
                    one_member = next(iter(enum["members"]))
                    rte_translation = '\n'.join(
                        f"#define {member.upper()} {member}" for member in enum["members"].keys()
                    )
                    enum_definition = (
                        f'\n'
                        f'#ifndef {one_member}\n'
                        f'{enum_definition}\n'
                        f'#else\n'
                        f'{rte_translation}\n'
                        f'#endif /* {one_member} */\n'
                    )
                fh.write(enum_definition)
            fh.write(f'#endif /* {guard} */\n')

    def generate_common_header_files(self):
        """ Generates common header files.

        These headers can be included in general files such as VcDummy_spm.h or VcExtVar.h.

        Returns:
            common_header_files (list(Path)): List of names of generated files.
        """
        src_dir = self._build_prj_cfg.get_src_code_dst_dir()
        enum_file_path = Path(src_dir, 'VcEnumerations.h')
        struct_file_path = Path(src_dir, 'VcStructs.h')
        self._generate_enum_header_file(enum_file_path)
        self._generate_struct_header_file(struct_file_path)
        self.common_header_files.extend([enum_file_path.name, struct_file_path.name])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for handling the Service abstraction."""


import re

from powertrain_build.interface.base import filter_signals
from powertrain_build.interface.csp_api import CspApi
from powertrain_build.interface.application import get_internal_domain
from powertrain_build.lib import logger

LOGGER = logger.create_logger("service")


def get_service(app, client_name, interface):
    """Get service implementation specification"""
    rasters = app.get_rasters()
    LOGGER.debug("Rasters: %s", rasters)
    translation_files = app.get_translation_files()

    sfw = ServiceFramework(app)
    sfw.filter = f"{client_name}_internal"
    sfw.name = f"{client_name}_{interface}"
    sfw.parse_definition(translation_files)
    internal = get_internal_domain(rasters)
    properties_from_json = [
        {"destination": "min", "source": "min", "default": "-"},
        {"destination": "max", "source": "max", "default": "-"},
        {"destination": "variable_type", "source": "type"},
        {"destination": "offset", "source": "offset", "default": "-"},
        {"destination": "factor", "source": "lsb", "default": 1},
        {"destination": "description", "source": "description"},
        {"destination": "unit", "source": "unit", "default": "-"},
    ]
    for raster in rasters:
        external_signals = filter_signals(raster.insignals, internal)
        sfw.add_signals(
            external_signals,
            "insignals",
            properties_from_json,
        )
        sfw.add_signals(raster.outsignals, "outsignals", properties_from_json)
    return sfw.to_model(interface)


def get_service_list(app):
    """Get service list from app

    Args:
        app (Application): Pybuild project

    Returns:
        (str): a string containing the translated service list
    """
    translation_map = app.get_service_mapping()
    cmake = ''
    for proxy, service in translation_map.items():
        lib = re.sub('-', '_', service + '_lib' + proxy + '_service_proxy').upper()
        include = re.sub('-', '_', service + '_include_dir').upper()
        cmake += f"LIST(APPEND extra_libraries ${{{lib}}})\n"
        cmake += f"LIST(APPEND EXTRA_INCLUDE_DIRS ${{{include}}})\n"
    return cmake


class ServiceFramework(CspApi):
    """Service Framework abstraction layer"""

    def __repr__(self):
        """String representation of SWFL"""
        return (
            f"<SWFL {self.name}"
            f" app_side insignals: {len(self.signal_names['app']['insignals'])}"
            f" app_side outsignals: {len(self.signal_names['app']['outsignals'])}>"
        )

    def get_map_file(self):
        """Get service translation map file

        Returns:
            (Path): service translation map file
        """
        return self.base_application.get_services_file()

    def get_map(self):
        """Get service translation map

        Returns:
            (dict): service translation map
        """
        return self.base_application.get_service_mapping()

    def check_endpoints(self):
        """Check and crash if signal endpoint contains both produces and consumes signals."""
        endpoints = {}
        for signal_name, signal_specs in self.translations.items():
            if signal_name in self.signal_names["app"]['insignals']:
                consumed = True
            elif signal_name in self.signal_names["app"]['outsignals']:
                consumed = False
            else:
                continue
            for signal_spec in signal_specs:
                endpoint = signal_spec[self.position.endpoint.value]
                api = signal_spec[self.position.api.value]
                key = (api, endpoint)
                if key not in endpoints:
                    endpoints[key] = {
                        "consumed": consumed,
                        "signals": set()
                    }
                endpoints[key]["signals"].add(signal_name)
                assert consumed == endpoints[key]["consumed"], \
                    f"Signal endpoint {endpoint} for {api} contains both consumed and produced signals"

    def extract_endpoint_definitions(self, raw):
        """Extract endpoint definitions from yaml file.

        Args:
            raw (dict): Raw yaml file
        Returns:
            (dict): Endpoint definitions
        """
        self.parse_api_definitions(raw.get("service", {}))

    @staticmethod
    def extract_definition(definition):
        """Extract definition from yaml file.
        Returns the properties and methods for a service.

        Args:
            definition (dict): Definition from yaml file
        Returns:
            (dict): Specifications for a service
        """
        specifications = {}
        specifications['properties'] = definition.get('properties', [])
        specifications['methods'] = definition.get('methods', [])
        return specifications

    def to_model(self, client):
        """Method to generate dict to be saved as yaml
        Args:
            client (str): Name of the client in signal comm
        Returns:
            spec (dict): Signal model for using a service
        """
        properties, types = self.properties_service_model(client)
        descriptions = {
            'internal': {
                'brief': "Internal interface for associated application.",
                'full': "This interface should only be used by the associated application."
            },
            'external': {
                'brief': "External interface.",
                'full': "This interface should be used by modules wanting to interact with the associated application."
            },
            'observer': {
                'brief': "Read-only interface.",
                'full': "This interface can be used by anyone wanting information from the associated application."
            },
        }
        model = {"name": self.name,
                 "version": "${SERVICE_VERSION}",
                 "description": descriptions[client],
                 "properties": properties,
                 "types": types}
        return model

    def properties_service_model(self, client):
        """Generate properties and types for a service

        Args:
            client (str): Name of the client in signal comm

        Returns:
            (list): List of properties
            (list): List of types
        """
        properties = {}
        types = {}
        accessors = {}
        if client == 'internal':
            accessors['insignals'] = 'r-'
            accessors['outsignals'] = '-w'
        elif client == 'external':
            accessors['insignals'] = '-w'
            accessors['outsignals'] = 'r-'
        else:
            accessors['insignals'] = 'r-'
            accessors['outsignals'] = 'r-'

        properties_in, types_in = self._properties_service_model(
            self.signal_names["app"]["insignals"],
            accessors['insignals'])
        properties_out, types_out = self._properties_service_model(
            self.signal_names["app"]["outsignals"],
            accessors['outsignals'])

        properties = properties_in + properties_out
        types = types_in + types_out
        return properties, types

    def _specifications(self, signal_names):
        """ Iterate over signal specifications for allowed services

        Args:
            signal_names (list): allowed signals
        Yields:
            specification (dict): Specification for a signal for an allowed service
        """
        for _, spec in self._generator(signal_names, unique_names=False):
            yield spec

    def _properties_service_model(self, signal_names, accessors):
        """ Placeholder
        """
        properties = []
        endpoint_members = {}
        endpoint_types = {}
        for signal_spec in self._specifications(signal_names):
            interface = signal_spec[self.position.api.value]
            if self.skip_interface(interface):
                continue
            endpoint = signal_spec[self.position.endpoint.value]
            primitive = signal_spec[self.position.property_name.value]
            if endpoint not in endpoint_members and primitive is not None:
                endpoint_members[endpoint] = []
            if primitive is not None:
                if endpoint not in endpoint_types:
                    endpoint_types[endpoint] = {
                        'name': endpoint,
                        'kind': 'struct',
                        'description': {
                            'brief': endpoint,
                            "full": "Generated from project without custom description"
                        },
                        'members': []
                    }
                endpoint_members[endpoint].append({
                    'name': primitive,
                    'type': primitive,
                })
                endpoint_types[endpoint]['members'].append({
                    'name': primitive,
                    'type': signal_spec[self.position.property_type.value],
                })
            else:
                primitive_type = signal_spec[self.position.property_type.value]
                primitive_desc = signal_spec[self.position.description.value]
                primitive_unit = signal_spec[self.position.unit.value]
                properties.append(
                    {
                        'name': endpoint,
                        'type': primitive_type,
                        'unit': primitive_unit,
                        'accessors': accessors,
                        'description': {
                            'brief': endpoint,
                            'full': primitive_desc,
                        }
                    }
                )
        for endpoint_name in sorted(endpoint_members):
            properties.append(
                {
                    'name': endpoint_name,
                    'type': endpoint_name,
                    'unit': 'struct',
                    'accessors': accessors,
                    'description': {
                        'brief': endpoint_name,
                        "full": "Generated from project without custom description"},
                }
            )
        return_types = []
        for endpoint_name, endpoint_type in endpoint_types.items():
            return_types.append(endpoint_type)
        return properties, return_types

    def skip_interface(self, interface):
        """ Filter services not in list.

        Args:
            service (str): interface
        Returns:
            skip (bool): Skip this interface
        """
        if self.filter is None:
            LOGGER.debug('No interface filter. Allowing everyone.')
            return False
        if interface in self.filter:
            LOGGER.debug('%s is in %s', interface, self.filter)
            return False
        return True



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module containing cvc classes for VCC - defines and includes for memory sections."""

import time
from pathlib import Path
from powertrain_build import build_defs
from powertrain_build.problem_logger import ProblemLogger


class MemorySection(ProblemLogger):
    """Handle headers for CVC_* definitions."""

    calibration_definitions = [
        'CVC_CAL',
        'CVC_CAL_ASIL_A',
        'CVC_CAL_ASIL_B',
        'CVC_CAL_ASIL_C',
        'CVC_CAL_ASIL_D',
        'CVC_CAL_MERGEABLE_ASIL_A',
        'CVC_CAL_MERGEABLE_ASIL_B',
        'CVC_CAL_MERGEABLE_ASIL_C',
        'CVC_CAL_MERGEABLE_ASIL_D'
    ]
    measurable_definitions = [
        'CVC_DISP',
        'CVC_DISP_ASIL_A',
        'CVC_DISP_ASIL_B',
        'CVC_DISP_ASIL_C',
        'CVC_DISP_ASIL_D'
    ]

    def __init__(self, build_cfg, unit_cfg):
        super().__init__()
        self.build_cfg = build_cfg
        self.unit_cfg = unit_cfg
        self.mem_map_config = self.build_cfg.get_memory_map_config()
        self.include_header_guards = self.mem_map_config['includeHeaderGuards']
        self.mem_map_include = f'#include "{self.mem_map_config["memMapPrefix"]}_MemMap.h"\n'
        self.xcp_enabled = self.build_cfg.get_xcp_enabled()
        self.use_volatile_globals = self.build_cfg.get_use_volatile_globals()

    @staticmethod
    def _get_mem_map_section(section):
        return 'STOP' if section == 'END' else section

    @staticmethod
    def _get_header(section_file):
        section_file_header_guard = section_file.split('.')[0].upper()
        return [
            f'#ifndef {section_file_header_guard}_H\n',
            f'#define {section_file_header_guard}_H\n\n'
        ]

    @staticmethod
    def _get_footer(section_file):
        section_file_header_guard = section_file.split('.')[0].upper()
        return [f'\n#endif /* {section_file_header_guard}_H */\n']

    def _get_calibration_rte_macro_expansion(self, section_file):
        macros = self._get_header(section_file)
        swc_name = self.build_cfg.get_composition_config("softwareComponentName")
        macros.append(f'#include "Rte_{swc_name}.h"\n')

        config = self.unit_cfg.get_per_cfg_unit_cfg()
        valid_configs = ["outports", "local_vars", "calib_consts"]
        for valid_config in valid_configs:
            for signal_name, unit_info in config.get(valid_config, {}).items():
                define_str = f'#define {signal_name} Rte_CData_{swc_name}_{signal_name}()'
                if signal_name.startswith("m") and not signal_name.endswith("_r") and not signal_name.endswith("_c"):
                    define_str += f"->dt_{signal_name}"
                define_str += "\n"
                for info in unit_info.values():  # Should be length 1 for cal
                    if "CVC_CAL" in info["class"]:
                        macros.append(define_str)

        macros.extend(self._get_footer(section_file))

        return macros

    def _get_cal(self, section):
        cvc_undefines = [f'#undef {definition}\n' for definition in self.calibration_definitions]
        if section == 'START':
            volatile_string = 'volatile' if self.use_volatile_globals else ''
            cvc_defines = [f'#define {definition} {volatile_string}\n' for definition in self.calibration_definitions]
        else:
            cvc_defines = []
        section_type = 'cal' if self.xcp_enabled else 'disp'
        memory_section_handling = [
            self.mem_map_config['projectDefines'][self._get_mem_map_section(section)][section_type] + '\n'
        ]
        if self.mem_map_config['includeMemMapForCalibration'] or not self.xcp_enabled:
            memory_section_handling.append(self.mem_map_include)
        return cvc_undefines, cvc_defines, memory_section_handling

    def _get_disp(self, section):
        cvc_undefines = [f'#undef {definition}\n' for definition in self.measurable_definitions]
        if section == 'START':
            volatile_string = 'volatile' if self.use_volatile_globals else ''
            cvc_defines = [f'#define {definition} {volatile_string}\n' for definition in self.measurable_definitions]
        else:
            cvc_defines = []
        memory_section_handling = [
            self.mem_map_config['projectDefines'][self._get_mem_map_section(section)]['disp'] + '\n',
            self.mem_map_include
        ]
        return cvc_undefines, cvc_defines, memory_section_handling

    def _get_code(self, section):
        cvc_undefines = []
        cvc_defines = []
        memory_section_handling = [
            self.mem_map_config['projectDefines'][self._get_mem_map_section(section)]['code'] + '\n',
            self.mem_map_include
        ]
        return cvc_undefines, cvc_defines, memory_section_handling

    def _get_const(self, section):
        cvc_undefines = []
        cvc_defines = []
        memory_section_handling = [
            self.mem_map_config['projectDefines'][self._get_mem_map_section(section)]['const'] + '\n',
            self.mem_map_include
        ]
        return cvc_undefines, cvc_defines, memory_section_handling

    def _get_nvm(self, section):
        cvc_undefines = []
        cvc_defines = []
        memory_section_handling = [
            self.mem_map_config['projectDefines'][self._get_mem_map_section(section)]['nvm'] + '\n',
            self.mem_map_include
        ]
        return cvc_undefines, cvc_defines, memory_section_handling

    def _get_rest(self):
        cvc_undefines = []
        cvc_defines = []
        memory_section_handling = [
            self.mem_map_include
        ]
        return cvc_undefines, cvc_defines, memory_section_handling

    def _get_predecl(self):
        cvc_undefines = []
        cvc_defines = []
        memory_section_handling = []
        return cvc_undefines, cvc_defines, memory_section_handling

    def generate_cvc_header(self, section, section_file):
        """Generate CVC headers.

        Args:
            section (str): Name of the CVC section
            section_file (str): Name of the header file
        Returns:
            lines_to_write (list(str)): Lines to write to given section file.
        """
        header = self._get_header(section_file) if self.include_header_guards else []
        footer = self._get_footer(section_file) if self.include_header_guards else []
        if '_CAL_' in section_file:
            cvc_undefines, cvc_defines, memory_section_handling = self._get_cal(section)
        elif '_DISP_' in section_file:
            cvc_undefines, cvc_defines, memory_section_handling = self._get_disp(section)
        elif not section_file.startswith('PREDECL_'):
            if section_file.startswith('CVC_CAL') or section_file.startswith('CVC_DISP'):
                self.critical('Should not find CVC_CAL/DISP here. Check logic. File: %s.', section_file)
            elif section_file.startswith('CVC_CODE'):
                cvc_undefines, cvc_defines, memory_section_handling = self._get_code(section)
                use_rte_macro_expansion = self.build_cfg.get_code_generation_config("useCalibrationRteMacroExpansion")
                if section == 'START' and use_rte_macro_expansion:
                    memory_section_handling.extend(self._get_calibration_rte_macro_expansion(section_file))
                    # header and footer are part of the extension
                    header = []
                    footer = []
            elif section_file.startswith('CVC_CONST'):
                cvc_undefines, cvc_defines, memory_section_handling = self._get_const(section)
            else:
                cvc_undefines, cvc_defines, memory_section_handling = self._get_rest()
        else:
            cvc_undefines, cvc_defines, memory_section_handling = self._get_predecl()

        return header + cvc_undefines + cvc_defines + memory_section_handling + footer

    def generate_required_header_files(self):
        """Generate required header files to delivery folder.

        Generate required header files such as memory protection files.
        NOTE: Currently, only one ASIL level can be selected for an SWC.
        """
        self.info('******************************************************')
        self.info('Start generating required header files')
        start_time = time.time()
        src_dst_dir = self.build_cfg.get_src_code_dst_dir()
        for section_dict in build_defs.PREDECL_EXTRA.values():
            for section_file in section_dict.values():
                header = self._get_header(section_file) if self.include_header_guards else []
                footer = self._get_footer(section_file) if self.include_header_guards else []
                with Path(src_dst_dir, section_file).open('w', encoding="utf-8") as header_file_handler:
                    header_file_handler.writelines(header + footer)

        for asil_dict in build_defs.ASIL_LEVEL_MAP.values():
            for type_dict in asil_dict.values():
                for section_dict in type_dict.values():
                    for section, section_file in section_dict.items():
                        lines_to_write = self.generate_cvc_header(section, section_file)
                        with Path(src_dst_dir, section_file).open('w', encoding="utf-8") as header_file_handler:
                            header_file_handler.writelines(lines_to_write)

        for nvm_type, nvm_dict in build_defs.NVM_LEVEL_MAP.items():
            for section_dict in nvm_dict.values():
                for section, section_file in section_dict.items():
                    section_config = self.mem_map_config['projectDefines'][self._get_mem_map_section(section)]
                    if "nvm" in section_config:
                        lines_to_write = self._get_nvm(section)[2]
                    else:
                        lines_to_write = []
                    header = self._get_header(section_file) if self.include_header_guards else []
                    footer = self._get_footer(section_file) if self.include_header_guards else []
                    with Path(src_dst_dir, section_file).open('w', encoding="utf-8") as header_file_handler:
                        header_file_handler.writelines(header + lines_to_write + footer)
        self.info('Finished generating required header files (in %4.2f s)', time.time() - start_time)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module to create an a2l file from a conversion table file."""

import argparse
import json
import sys
from pathlib import Path
from typing import List, Optional


PARSER_HELP = "Create a2l file from conversion_table.json file."


def get_vtab_text(vtab):
    """Convert vtab dict to a2l text."""

    vtab_text = (
        '    /begin COMPU_VTAB\n'
        f'        CONV_TAB_{vtab["name"]}             /* Name */\n'
        '        "Conversion table"          /* LongIdentifier */\n'
        '        TAB_VERB            /* ConversionType */\n'
        f'        {len(vtab["disp_values"])}          /* NumberValuePairs */\n'
    )

    vtab_text += ''.join(
        f'        {vtab["start_value"]+i}          /* InVal */\n'
        f'        "{value}"          /* OutVal */\n'
        for i, value in enumerate(vtab['disp_values'])
    )

    vtab_text += '    /end COMPU_VTAB\n\n'

    return vtab_text


def create_conversion_table(input_json: Path, output_a2l: Path):
    """Create a2l conversion table for custom units."""
    with open(input_json.resolve(), encoding="utf-8") as f_h:
        conversion_table = json.load(f_h)

    with open(output_a2l.resolve(), 'w', encoding="utf-8") as f_h:
        for vtab in conversion_table:
            f_h.write(get_vtab_text(vtab))


def create_conversion_table_cli(args: argparse.Namespace):
    """CLI wrapper function for passing in Namespace object.

    This allows maintaining a standardized CLI function signature while not breaking backwards
    compatibility with create_converstion_table.
    """
    create_conversion_table(args.input_file, args.output_file)


def configure_parser(parser: argparse.ArgumentParser):
    """Set up parser for CLI."""
    parser.add_argument('input_file', type=Path)
    parser.add_argument('output_file', type=Path)
    parser.set_defaults(func=create_conversion_table_cli)


def main(argv: Optional[List[str]] = None):
    """Main function for CLI."""
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


if __name__ == '__main__':
    main(sys.argv[1:])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

from argparse import ArgumentParser, Namespace
from typing import List, Optional

import powertrain_build.check_interface
import powertrain_build.create_conversion_table
import powertrain_build.interface.export_global_vars
import powertrain_build.interface.generate_adapters
import powertrain_build.interface.generate_hi_interface
import powertrain_build.interface.generate_service
import powertrain_build.interface.generate_wrappers
import powertrain_build.interface.model_yaml_verification
import powertrain_build.interface.update_model_yaml
import powertrain_build.interface.update_call_sources
import powertrain_build.replace_compu_tab_ref
import powertrain_build.signal_inconsistency_check
from powertrain_build import __version__
from powertrain_build.config import ProcessHandler
from powertrain_build.lib import logger
from powertrain_build.wrapper import PyBuildWrapper

LOGGER = logger.create_logger(__file__)


def parse_args(argv: Optional[List[str]] = None) -> Namespace:
    """Parse command line arguments."""
    parser = ArgumentParser(
        prog="powertrain-build",
        description="Powertrain-build",
    )
    parser.add_argument(
        "-V", "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    command_subparsers = parser.add_subparsers(title="Commands", dest="command", required=True)

    wrapper_parser = command_subparsers.add_parser(
        "wrapper",
        help=PyBuildWrapper.PARSER_HELP,
    )
    PyBuildWrapper.add_args(wrapper_parser)

    config_parser = command_subparsers.add_parser(
        "config",
        help=ProcessHandler.PARSER_HELP,
    )
    ProcessHandler.configure_parser(config_parser)

    check_interface_parser = command_subparsers.add_parser(
        "check-interface",
        help=powertrain_build.check_interface.PARSER_HELP,
    )
    powertrain_build.check_interface.configure_parser(check_interface_parser)

    create_conversion_table_parser = command_subparsers.add_parser(
        "create-conversion-table",
        help=powertrain_build.create_conversion_table.PARSER_HELP,
    )
    powertrain_build.create_conversion_table.configure_parser(create_conversion_table_parser)

    replace_compu_tab_ref_parser = command_subparsers.add_parser(
        "replace-compu-tab-ref",
        help=powertrain_build.replace_compu_tab_ref.PARSER_HELP,
    )
    powertrain_build.replace_compu_tab_ref.configure_parser(replace_compu_tab_ref_parser)

    signal_inconsistency_check_parser = command_subparsers.add_parser(
        "signal-inconsistency-check",
        help=powertrain_build.signal_inconsistency_check.PARSER_HELP,
    )
    powertrain_build.signal_inconsistency_check.configure_parser(signal_inconsistency_check_parser)

    interface_parser = command_subparsers.add_parser(
        "interface",
        help="Interface commands",
    )
    interface_subparsers = interface_parser.add_subparsers(
        title="Interface commands",
        dest="interface_command",
        required=True,
    )

    export_global_vars_parser = interface_subparsers.add_parser(
        "export-global-vars",
        help=powertrain_build.interface.export_global_vars.PARSER_HELP,
    )
    powertrain_build.interface.export_global_vars.configure_parser(export_global_vars_parser)

    generate_adapters_parser = interface_subparsers.add_parser(
        "generate-adapters",
        help=powertrain_build.interface.generate_adapters.PARSER_HELP,
    )
    powertrain_build.interface.generate_adapters.configure_parser(generate_adapters_parser)

    generate_hi_interface_parser = interface_subparsers.add_parser(
        "generate-hi-interface",
        help=powertrain_build.interface.generate_hi_interface.PARSER_HELP,
    )
    powertrain_build.interface.generate_hi_interface.configure_parser(generate_hi_interface_parser)

    generate_service_parser = interface_subparsers.add_parser(
        "generate-service",
        help=powertrain_build.interface.generate_service.PARSER_HELP,
    )
    powertrain_build.interface.generate_service.configure_parser(generate_service_parser)

    generate_wrappers_parser = interface_subparsers.add_parser(
        "generate-wrappers",
        help=powertrain_build.interface.generate_wrappers.PARSER_HELP,
    )
    powertrain_build.interface.generate_wrappers.configure_parser(generate_wrappers_parser)

    model_yaml_verification_parser = interface_subparsers.add_parser(
        "model-yaml-verification",
        help=powertrain_build.interface.model_yaml_verification.PARSER_HELP,
    )
    powertrain_build.interface.model_yaml_verification.configure_parser(model_yaml_verification_parser)

    update_model_yaml_parser = interface_subparsers.add_parser(
        "update-model-yaml",
        help=powertrain_build.interface.update_model_yaml.PARSER_HELP,
    )
    powertrain_build.interface.update_model_yaml.configure_parser(update_model_yaml_parser)

    update_call_sources_parser = interface_subparsers.add_parser(
        "update-call-sources",
        help=powertrain_build.interface.update_call_sources.PARSER_HELP,
    )
    powertrain_build.interface.update_call_sources.configure_parser(update_call_sources_parser)

    return parser.parse_args(argv)


def main(argv: Optional[List[str]] = None) -> Namespace:
    """Run main function."""
    args = parse_args(argv)
    return args.func(args)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Python module used for building a Vcc SPM SW release.

This is the entry point to powertrain-build which includes all other needed modules.
Loads configuration files and sequences the code generation steps.
"""

import glob
import logging
import os
import shutil
import sys
import time
from os.path import join as pjoin
from pathlib import Path

from powertrain_build import __config_version__, __version__, build_defs
from powertrain_build.a2l_merge import A2lMerge
from powertrain_build.build_proj_config import BuildProjConfig
from powertrain_build.core import Core, HICore, ZCCore
from powertrain_build.core_dummy import CoreDummy
from powertrain_build.create_conversion_table import create_conversion_table
from powertrain_build.dids import DIDs, HIDIDs, ZCDIDs
from powertrain_build.dummy import DummyVar
from powertrain_build.dummy_spm import DummySpm
from powertrain_build.ext_dbg import ExtDbg
from powertrain_build.ext_var import ExtVarCsv, ExtVarYaml
from powertrain_build.feature_configs import FeatureConfigs
from powertrain_build.lib.helper_functions import create_dir, get_repo_root, merge_dicts
from powertrain_build.memory_section import MemorySection
from powertrain_build.nvm_def import NVMDef, ZCNVMDef
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.replace_compu_tab_ref import replace_tab_verb
from powertrain_build.rte_dummy import RteDummy
from powertrain_build.sched_funcs import SchedFuncs
from powertrain_build.signal_if_html_rep import SigIfHtmlReport
from powertrain_build.signal_incons_html_rep import SigConsHtmlReport
from powertrain_build.signal_interfaces import CsvSignalInterfaces, YamlSignalInterfaces
from powertrain_build.unit_configs import CodeGenerators, UnitConfigs
from powertrain_build.user_defined_types import UserDefinedTypes
from powertrain_build.zone_controller.calibration import ZoneControllerCalibration
from powertrain_build.zone_controller.composition_yaml import CompositionYaml

LOG = logging.getLogger()
REPO_ROOT = get_repo_root()


def setup_logging(log_dst_dir, problem_logger, debug=True, quiet=False):
    """Set up the python logger for the build environment.

    Three logger streams are set up. One that log to stdout (info level),
    and one to a build.log file (info level), and finally, on stream that
    logs to build_dbg.log (debug level log). The log files are put in the
    directory configured in the config file.

    Args:
        log_dst_dir (str): the path to where the log file should be stored
        problem_logger (obj): the ProblemLogger object to initialise
        debug (bool): True - if debug log shall be generated
        quiet (bool): False - disable logging to stdout

    """
    LOG.setLevel(logging.DEBUG)
    LOG.handlers = []  # Remove all previous loggers
    logging.captureWarnings(True)
    # Setup debug build logger
    log_file = pjoin(log_dst_dir, "build_dbg.log")
    if debug:
        dbg = logging.FileHandler(log_file)
        dbg.setLevel(logging.DEBUG)
        dbg_formatter = logging.Formatter(
            "%(asctime)s - %(module)s."
            "%(funcName)s [%(lineno)d]"
            " - %(levelname)s - %(message)s"
        )
        dbg.setFormatter(dbg_formatter)
        LOG.addHandler(dbg)
    # Setup normal build logger
    log_file = pjoin(log_dst_dir, "build.log")
    nrm = logging.FileHandler(log_file)
    nrm.setLevel(logging.INFO)
    build_log_frmt = logging.Formatter("%(asctime)s - %(levelname)s" " - %(message)s")
    nrm.setFormatter(build_log_frmt)
    LOG.addHandler(nrm)

    if not quiet:
        nrm_strm = logging.StreamHandler(sys.stdout)
        nrm_strm.setLevel(logging.INFO)
        nrm_strm.setFormatter(build_log_frmt)
        LOG.addHandler(nrm_strm)

        error_strm = logging.StreamHandler(sys.stderr)
        error_strm.setLevel(logging.CRITICAL)
        error_strm.setFormatter(build_log_frmt)
        LOG.addHandler(error_strm)

    problem_logger.init_logger(LOG)


def check_interfaces(build_cfg, signal_if):
    """Check the interfaces.

    Checks interfaces in all configurations, and generates a html
    report with the result of the checks.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where
            files should be stored
        signal_if (SignalInterfaces): class holding signal interface information

    """
    LOG.info("******************************************************")
    LOG.info("Start check interface inconsistencies")
    start_time = time.time()
    report_dst_dir = build_cfg.get_reports_dst_dir()
    signal_inconsistency_result = signal_if.check_config()
    sig_report = SigConsHtmlReport(signal_inconsistency_result)
    LOG.info(
        "Finished check interface inconsistencies (in %4.2f s)",
        time.time() - start_time,
    )

    start_time = time.time()
    LOG.info("******************************************************")
    LOG.info("Start generating interface inconsistencies html-report")
    sig_report.generate_report_file(pjoin(report_dst_dir, "SigCheck.html"))
    LOG.info(
        "Finished - generating interface inconsistencies html-report (in %4.2f s)",
        time.time() - start_time,
    )


def interface_report(build_cfg, unit_cfg, signal_if):
    """Create report of signal interfaces.

    Creates a report of interfaces in all configurations, and generates
    a html-report with the result of the checks.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where
            files should be stored.
        unit_cfg (UnitConfigs): Class holding all unit interfaces.
        signal_if (SignalInterfaces): class holding signal interface information.

    """
    LOG.info("******************************************************")
    LOG.info("Start creating interface report")
    start_time = time.time()
    report_dst_dir = build_cfg.get_reports_dst_dir()

    sig_if = SigIfHtmlReport(build_cfg, unit_cfg, signal_if)
    sig_if.generate_report_file(pjoin(report_dst_dir, "SigIf.html"))
    LOG.info(
        "Finished - create interface report (in %4.2f s)", time.time() - start_time
    )


def generate_dummy_spm(build_cfg, unit_cfg, feature_cfg, signal_if, udt):
    """Generate c-files that define unit output signals.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where files should be stored.
        unit_cfg (UnitConfigs): Aggregated unit configs class.
        feature_cfg (FeatureConfigs): Used as a library to generate C-macros.
        signal_if (SignalInterfaces): Class holding signal interface information.
        udt (UserDefinedTypes): Class holding user defined data types.
    """

    def _add_undefined_signals_from_unit(data):
        """Add undefined signals from unit.

        Includes included configs.
        Arguments:
        data (dict): Data for the unit
        """
        for signal_name, inport_attributes in data.get("inports", {}).items():
            if signal_name in defined_ports:
                continue
            if (
                feature_cfg.check_if_active_in_config(inport_attributes["configs"])
                or not build_cfg.allow_undefined_unused
            ):
                defined_ports.append(signal_name)
                undefined_outports.append(inport_attributes)
        for include_unit in data.get("includes", []):
            include_data = unit_cfg.get_unit_config(include_unit)
            _add_undefined_signals_from_unit(include_data)

    LOG.info("******************************************************")
    LOG.info("Start generating output vars")
    start_time = time.time()
    dst_dir = build_cfg.get_src_code_dst_dir()
    # Get out ports for all units in project regardless of code switches
    unit_vars = unit_cfg.get_per_unit_cfg_total()
    undefined_outports = []
    defined_ports = []

    for data in unit_vars.values():
        for outport_name, outport_data in data.get("outports", {}).items():
            if not feature_cfg.check_if_active_in_config(outport_data["configs"]):
                LOG.debug("Outport %s not active in current project", outport_name)
            elif outport_name not in defined_ports:
                if not outport_data.get("class").startswith("CVC_EXT"):
                    defined_ports.append(outport_name)
    defined_ports.extend(signal_if.get_externally_defined_ports())

    for data in unit_vars.values():
        _add_undefined_signals_from_unit(data)

    dummy_spm = DummySpm(
        undefined_outports, build_cfg, feature_cfg, unit_cfg, udt, "VcDummy_spm"
    )
    dummy_spm.generate_files(dst_dir)
    LOG.info("Finished generating output vars (in %4.2f s)", time.time() - start_time)


def generate_did_files(build_cfg, unit_cfg):
    """Generate DIDAPI definition files.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where
            files should be stored
        unit_cfg (UnitConfigs): class holding units definitions,
            and which units to include

    """
    start_time = time.time()
    did_defs_files = pjoin(build_cfg.get_src_code_dst_dir(), "VcDIDDefinition")
    car_com_file = build_cfg.get_car_com_dst()
    LOG.info("******************************************************")
    LOG.info("Start generating %s.c&h", did_defs_files)
    dids = DIDs(build_cfg, unit_cfg)
    dids.gen_did_def_files(did_defs_files)
    LOG.info(
        "Finished generating %s.c&h  (in %4.2f s)",
        did_defs_files,
        time.time() - start_time,
    )
    LOG.info("******************************************************")
    LOG.info("Start generating %s", car_com_file)
    start_time = time.time()
    dids.gen_did_carcom_extract(car_com_file)
    LOG.info(
        "Finished generating %s  (in %4.2f s)", car_com_file, time.time() - start_time
    )
    LOG.info("******************************************************")


def generate_core_dummy(build_cfg, core, unit_cfg):
    """Generate the Core dummy files.

    The core dummy creates RTE dummy functions,
    and Id variables for enabling testing of VCC Software. If this dummy is not
    included, it is not possible to build the project until the supplier deliver
    an updated diagnostic core SW.

    Note:
        These dummy files shall not be delivered to the supplier!

    Args:
        build_cfg (BuildProjConfig): Build project class holding where
            files should be stored
        core (Core): class holding core configuration information

    """
    core_dummy_fname = build_cfg.get_core_dummy_name()
    LOG.info("******************************************************")
    LOG.info("Start generating Core Dummy - %s", core_dummy_fname)
    start_time = time.time()
    core_dummy = CoreDummy(core.get_current_core_config(), unit_cfg)
    ecu_supplier = build_cfg.get_ecu_info()[0]
    if ecu_supplier in ["Denso", "CSP"]:
        core_dummy.generate_dg2_core_dummy_files(core_dummy_fname)
    elif ecu_supplier == "RB":
        core_dummy.generate_rb_core_dummy_files(core_dummy_fname)
    else:
        msg = f"Could not generate VcCoreDummy, cannot identify the supplier {ecu_supplier}."
        LOG.critical(msg)
        raise ValueError(msg)
    LOG.info("Finished generating Core Dummy (in %4.2f s)", time.time() - start_time)


def generate_external_var(build_cfg, unit_cfg, udt, asil_level_dep, nrm_dict, dep_dict, sec_dict):
    """Generate two c-files that define the signal interface to the supplier.

    The VcExtVar function assigns all variables to the CVC_DISP memory area,
    while the ExtVarSafe.c are allocated to the CVC_DISP_ASIL_B memory area.
    Note that this function only declares the variables
    in the supplier interface, which the supplier writes to. All other
    variables shall be declared by the function which writes to the variable.
    Note that dummy functionality/variables should be created in the function
    to keep the signalling interface consistent!

    Args:
        build_cfg (BuildProjConfig): Build project class holding where files should be stored.
        unit_cfg (UnitConfigs) : Aggregated unit configs class.
        udt (UserDefinedTypes): Class holding user defined data types.
        asil_level_dep (str): ASIL level for the dependability variables.
        nrm_dict (dict): Dictionary with normal variables.
        dep_dict (dict): Dictionary with dependability variables.
        sec_dict (dict): Dictionary with secure variables.
    """
    if build_cfg.has_yaml_interface:
        ext_var_nrm = ExtVarYaml(nrm_dict, build_cfg, unit_cfg, udt)
        ext_var_dep = ExtVarYaml(dep_dict, build_cfg, unit_cfg, udt, asil_level_dep)
    else:
        ext_var_nrm = ExtVarCsv(nrm_dict, build_cfg, unit_cfg, udt)
        ext_var_dep = ExtVarCsv(dep_dict, build_cfg, unit_cfg, udt, asil_level_dep)
        ext_var_sec = ExtVarCsv(sec_dict, build_cfg, unit_cfg, udt, build_defs.SECURE)

    ext_var_instances = {
        ext_var_nrm: "VcExtVar",
        ext_var_dep: "VcExtVarSafe",
    }

    if not build_cfg.has_yaml_interface:
        ext_var_instances[ext_var_sec] = "VcExtVarSecure"

    for instance, dir_name in ext_var_instances.items():
        ext_var_path = Path(build_cfg.get_src_code_dst_dir(), dir_name)
        instance.generate_files(ext_var_path)


def generate_dummy_var(build_cfg, unit_cfg, signal_if, udt):
    """Generate c-file that define the missing signals.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where files should be stored.
        unit_cfg (UnitConfigs) : Aggregated unit configs class.
        signal_if (SignalInterfaces): class holding signal interface information.
        udt (UserDefinedTypes): Class holding user defined data types.
    """
    LOG.info("******************************************************")
    LOG.info("Start generating VcDummy")
    start_time = time.time()
    dst_dir = build_cfg.get_src_code_dst_dir()
    nrm_dict, dep_dict, sec_dict, _ = signal_if.get_external_io()
    nrm_dict = merge_dicts(nrm_dict, dep_dict, merge_recursively=True)
    nrm_dict = merge_dicts(nrm_dict, sec_dict, merge_recursively=True)
    res_dict = signal_if.check_config()
    dummy = DummyVar(unit_cfg, nrm_dict, res_dict, build_cfg, udt)
    dummy.generate_file(pjoin(dst_dir, "VcDummy"))
    LOG.info("Finished generating VcDummy (in %4.2f s)", time.time() - start_time)


def generate_nvm_def(build_cfg, unit_cfg, no_nvm_a2l):
    """Generate the c&h-files which declares the NVM-ram.

    The NVM-ram is declared in a struct per datatype length, in order
    to provide a defined order in RAM/FLASH/EEPROM for the tester
    communication service. Furthermore, # defines with the variables are
    created to minimize the needed model changes for access to the memory.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where files should be stored.
        unit_cfg (UnitConfigs): class holding units definitions, and which units to include.
        no_nvm_a2l (bool): Do not generate A2L for NVM structs.
    """
    LOG.info("******************************************************")
    LOG.info("Start generating NVMDefinitions")
    start_time = time.time()
    tot_vars_nvm = unit_cfg.get_per_cfg_unit_cfg().get("nvm", {})
    nvm_def = NVMDef(build_cfg, unit_cfg, tot_vars_nvm)
    nvm_def.generate_nvm_config_files(no_nvm_a2l)
    LOG.info(
        "Finished generating NVMDefinitions (in %4.2f s)", time.time() - start_time
    )


def copy_unit_src_to_src_out(build_cfg):
    """Copy unit source code to delivery folder.

    Function to copy all relevant .c, .h and .a2l files to the src
    delivery folder (defined in the config file for the project),
    from the units that are included in the project.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where
            files should be stored
    """
    LOG.info("******************************************************")
    LOG.info("Start copying unit source files")
    start_time = time.time()
    src_dirs = build_cfg.get_unit_src_dirs()
    src_dst_dir = build_cfg.get_src_code_dst_dir()
    files = []
    for src_dir in src_dirs.values():
        files.extend(glob.glob(pjoin(src_dir, "*.c")))
        files.extend(glob.glob(pjoin(src_dir, "*.cpp")))
        files.extend(glob.glob(pjoin(src_dir, "*.h")))
    for file_ in files:
        shutil.copy2(file_, src_dst_dir)
        LOG.debug("copied %s to %s", file_, src_dst_dir)
    LOG.info(
        "Finished copying unit source files (in %4.2f s)", time.time() - start_time
    )


def copy_common_src_to_src_out(build_cfg):
    """Copy source code to delivery folder.

    Function to copy all relevant .c and .h files to the src
    delivery folder (defined in the config file for the project),
    from the units that are included in the project.

    Optionally, also patch the defined functions with the "schedulerPrefix".

    Args:
        build_cfg (BuildProjConfig): Build project class holding where files should be stored.
    """
    LOG.info("******************************************************")
    prefix = build_cfg.get_scheduler_prefix()
    if prefix:
        LOG.info("Start copying and patching common source files")
    else:
        LOG.info("Start copying common source files")

    start_time = time.time()
    src_dir = build_cfg.get_common_src_dir()
    src_dst_dir = build_cfg.get_src_code_dst_dir()
    included_common_files = build_cfg.get_included_common_files()

    files = [
        pjoin(src_dir, common_file + ".c") for common_file in included_common_files
    ]
    files.extend(
        [pjoin(src_dir, common_file + ".h") for common_file in included_common_files]
    )

    files_to_copy = filter(os.path.isfile, files)
    for file_ in files_to_copy:
        if prefix:
            patch_and_copy_common_src_to_src_out(prefix, Path(file_), Path(src_dst_dir))
        else:
            shutil.copy2(file_, src_dst_dir)
        LOG.debug("copied %s to %s", file_, src_dst_dir)

    LOG.info(
        "Finished copying common source files (in %4.2f s)", time.time() - start_time
    )


def patch_and_copy_common_src_to_src_out(prefix, file_path, dest_dir):
    """Copy common source code to delivery folder, patched with "schedulerPrefix" as prefix.

    Args:
        prefix (str): Prefix.
        file_path (Path): Path to file to patch and copy.
        dest_dir (Path): Destination directory for the patched file.
    """
    with file_path.open(mode="r", encoding="utf-8") as file_handle:
        content = file_handle.read()

    new_function = f"{prefix}{file_path.stem}"
    new_content = content.replace(f"{file_path.stem}(", f"{new_function}(")
    new_content_lines = new_content.splitlines()

    if file_path.suffix == ".h":
        defines_index = next(
            (idx for idx, line in enumerate(new_content_lines) if "DEFINES" in line and "(OPT)" not in line),
            None
        )
        if defines_index is not None:
            # Insert at defines_index + 2, +1 to get to the next line and +1 to get after multiline comment
            new_content_lines.insert(defines_index + 2, f"#define {file_path.stem} {new_function}")

    with Path(dest_dir, file_path.name).open(mode="w", encoding="utf-8") as file_handle:
        file_handle.write("\n".join(new_content_lines))


def copy_files_to_include(build_cfg):
    """Copy source code to delivery folder.

    Function to copy all extra include files to the src delivery folder
    (defined in the config file for the project),
    from the units that are included in the project.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where
            files should be stored
    """
    LOG.info("******************************************************")
    LOG.info("Start copying extra included source files")
    start_time = time.time()
    src_dst_dir = build_cfg.get_src_code_dst_dir()

    # Copy some files flat to the destination directory
    include_paths = build_cfg.get_includes_paths_flat()
    files = []
    for include_path in include_paths:
        if os.path.isdir(include_path):
            files.extend(
                [
                    Path(include_path, file_name)
                    for file_name in Path(include_path).iterdir()
                    if not file_name.is_dir()
                ]
            )
        elif os.path.isfile(include_path):
            files.append(include_path)
        else:
            LOG.critical("File or directory %s not found", include_path)
    for file_ in files:
        shutil.copy2(file_, src_dst_dir)
        LOG.debug("copied %s to %s", file_, src_dst_dir)

    # Copy some directories as is to the destination directory
    include_tree_paths = [Path(p) for p in build_cfg.get_includes_paths_tree()]
    existing_tree_paths = [p for p in include_tree_paths if p.exists()]
    missing_tree_paths = [p for p in include_tree_paths if not p.exists()]
    if missing_tree_paths:
        LOG.warning(f"Missing include paths: {missing_tree_paths}")
    for path in existing_tree_paths:
        for file_ in path.rglob("*"):
            if file_.is_file():
                new_path = Path(src_dst_dir, file_.relative_to(path))
                create_dir(new_path.parent)
                shutil.copy2(file_, new_path)
                LOG.debug("copied %s to %s", file_, new_path)
    LOG.info(
        "Finished copying extra included files (in %4.2f s)", time.time() - start_time
    )


def copy_unit_cfgs_to_output(build_cfg):
    """Copy all relevant unit configuration files to delivery folder.

    Function to copy all relevant unit config .json files to the UnitCfg
    delivery folder (defined in the config file for the project),
    from the units that are included in the project.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where
            files should be stored
    """
    cfg_dst_dir = build_cfg.get_unit_cfg_deliv_dir()
    if cfg_dst_dir is not None:
        LOG.info("******************************************************")
        LOG.info("Start copying the unit config files")
        start_time = time.time()
        cfg_dirs = build_cfg.get_unit_cfg_dirs()
        files = []
        for cfg_dir in cfg_dirs.values():
            files.extend(glob.glob(pjoin(cfg_dir, "*.json")))
        for file_ in files:
            shutil.copy2(file_, cfg_dst_dir)
            LOG.debug("copied %s to %s", file_, cfg_dst_dir)
        LOG.info(
            "Finished copying the unit config files (in %4.2f s)",
            time.time() - start_time,
        )


def merge_a2l_files(build_cfg, unit_cfg, complete_a2l=False, silver_a2l=False):
    """Merge a2l-files.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where
            files should be stored
        unit_cfg (UnitConfigs): class holding units definitions,
            and which units to include
    Returns:
        a2l (A2lMerge): A2lMerge class holding the merged a2l data.
    """
    LOG.info("******************************************************")
    LOG.info("Start merging A2L-files")
    start_time = time.time()
    src_dirs = build_cfg.get_unit_src_dirs()
    src_dst_dir = build_cfg.get_src_code_dst_dir()
    a2l_files_unit = []
    for src_dir in src_dirs.values():
        a2l_files_unit.extend(glob.glob(pjoin(src_dir, "*.a2l")))
    a2l_files_gen = glob.glob(pjoin(src_dst_dir, "*.a2l"))
    a2l = A2lMerge(build_cfg, unit_cfg, a2l_files_unit, a2l_files_gen)
    new_a2l_file = os.path.join(src_dst_dir, build_cfg.get_a2l_name())
    # LOG.debug(new_a2l_file)
    a2l.merge(new_a2l_file, complete_a2l, silver_a2l)
    LOG.info("Finished merging A2L-files (in %4.2f s)", time.time() - start_time)
    return a2l


def find_all_project_configs(prj_cfg_file):
    """Find all Project config files."""
    prj_root_dir, _ = os.path.split(prj_cfg_file)
    prj_root_dir = os.path.abspath(prj_root_dir)
    all_cfg_path = os.path.join(prj_root_dir, "..", "*", "ProjectCfg.json")
    return glob.glob(all_cfg_path, recursive=True)


def propagate_tag_name(build_cfg, tag_name, problem_logger):
    """Set tag name in relevant files, for release builds.

    Args:
        build_cfg (BuildProjConfig): Build project class holding where files should be stored.
        tag_name (str): git tag name.
        problem_logger (object): logger for powertrain_build.
    """
    LOG.info("******************************************************")
    LOG.info("Propagating tag name: %s", tag_name)
    start_time = time.time()

    src_dst_dir = build_cfg.get_src_code_dst_dir()
    h_file_path = os.path.join(src_dst_dir, "vcc_sp_version.h")

    if not os.path.isfile(h_file_path):
        problem_logger.critical("Missing %s", h_file_path)
        return

    with open(h_file_path, "r+", encoding="utf-8") as file_handle:
        contents = file_handle.read()
        file_handle.seek(0)
        new_contents = contents.replace(
            '#define VCC_SOFTWARE_NAME "tagname"',
            f'#define VCC_SOFTWARE_NAME "{tag_name}"',
        )
        file_handle.write(new_contents)

    LOG.info("Finished propagating tag name (in %4.2f s)", time.time() - start_time)


def add_args(parser):
    """Add command line arguments for powertrain_build.

    This is useful when powertrain-build should be run through a command line wrapper function.

    Args:
        parser (argparse.ArgumentParser): Parser instance to add arguments to.
    """
    powertrain_build_parser = parser.add_argument_group("powertrain-build arguments")
    powertrain_build_parser.add_argument(
        "--project-config", required=True, help="Project root configuration file"
    )
    powertrain_build_parser.add_argument(
        "--generate-system-info", action="store_true", help="Generate AllSystemInfo.mat"
    )
    powertrain_build_parser.add_argument(
        "--generate-custom-conv-tab",
        default=None,
        help="Path to conversion table file. Useful for TargetLink enums in A2L file.",
    )
    powertrain_build_parser.add_argument(
        "--core-dummy",
        action="store_true",
        help="Generate core dummy code to enable integration with old supplier code",
    )
    powertrain_build_parser.add_argument(
        "--rte-dummy",
        action="store_true",
        help="Generate RTE dummy code to enable e.g. Silver testing",
    )
    powertrain_build_parser.add_argument(
        "--debug", action="store_true", help="Activate the debug log"
    )
    powertrain_build_parser.add_argument(
        "--no-abort", action="store_true", help="Do not abort due to errors"
    )
    powertrain_build_parser.add_argument(
        "--no-nvm-a2l",
        action="store_true",
        help="Do not generate a2l file for NVM structs",
    )
    powertrain_build_parser.add_argument(
        "--complete-a2l", action="store_true", help="Generate A2L with project info"
    )
    powertrain_build_parser.add_argument(
        "--silver-a2l",
        action="store_true",
        help="Generate A2L file with Silver patching. Complete A2L argument takes precedence.",
    )
    powertrain_build_parser.add_argument(
        "--interface", action="store_true", help="Generate interface report"
    )
    powertrain_build_parser.add_argument(
        "--generate-rte-checkpoint-calls",
        action="store_true",
        help="Generate RTE function checkpoint calls",
    )
    powertrain_build_parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Display program version",
    )


def build(
    project_config,
    interface=False,
    core_dummy=False,
    rte_dummy=False,
    no_abort=False,
    no_nvm_a2l=False,
    debug=False,
    quiet=False,
    generate_system_info=False,
    generate_custom_conversion_table=None,
    complete_a2l=False,
    silver_a2l=False,
    generate_rte_checkpoint_calls=False,
):
    """Execute the build.

    Args:
        project_config (str): Project configuration file.
        interface (bool): Generate interface report. Default=False.
        core_dummy (bool): Generate core dummy code. Default=False.
        rte_dummy (bool): Generate RTE dummy code. Default=False.
        no_abort (bool): Do not abort due to errors. Default=False.
        no_nvm_a2l (bool): Do not generate A2L for NVM structs. Default=False.
        debug (bool): Activate the debug log. Default=False.
        quiet (bool): Disable logging to stdout. Default=False.
        generate_system_info (bool): Generate AllSystemInfo.mat for DocGen compatibility. Default=False.
        generate_custom_conversion_table (str): Path to conversion table file.
            Useful for TargetLink enums in A2L file. Default=None.
        complete_a2l (bool): Add an a2l header plus additional content such as XCP data.
        silver_a2l (bool): Add an a2l header plus additional patching required for Silver.
        generate_rte_checkpoint_calls (bool): Generate RTE function checkpoint calls.
    """
    try:
        problem_logger = ProblemLogger()
        tot_start_time = time.time()
        project_config_files = find_all_project_configs(project_config)
        prj_cfgs = {}

        for project_config_file in project_config_files:
            conf = BuildProjConfig(os.path.normpath(project_config_file))
            prj_cfgs.update({conf.name: conf})

        prj_cfgs = {}
        build_cfg = BuildProjConfig(os.path.normpath(project_config))
        prj_cfgs.update({build_cfg.name: build_cfg})

        build_cfg.create_build_dirs()
        src_dst_dir = build_cfg.get_src_code_dst_dir()

        setup_logging(build_cfg.get_log_dst_dir(), problem_logger, debug, quiet)
        LOG.info("Starting build")
        LOG.info("powertrain_build version is: %s", __version__)
        LOG.info("Project/Model config file version is: %s", __config_version__)
        LOG.info("Read SPM code switches")
        start_time = time.time()
        feature_cfg = FeatureConfigs(build_cfg)
        LOG.info(
            "Finished reading SPM code switches (in %4.2f s)", time.time() - start_time
        )

        LOG.info("******************************************************")
        LOG.info("Start generating per project unit config data")
        start_time = time.time()
        unit_cfg = UnitConfigs(build_cfg, feature_cfg)
        LOG.info(
            "Finished generating per project unit config data (in %4.2f s)",
            time.time() - start_time,
        )

        code_generation_config = build_cfg.get_code_generation_config()

        udt = UserDefinedTypes(build_cfg, unit_cfg)
        udt.generate_common_header_files()

        start_time = time.time()
        cnf_header = pjoin(src_dst_dir, build_cfg.get_feature_conf_header_name())
        LOG.info("******************************************************")
        LOG.info("Generate compiler switches header file %s", cnf_header)
        feature_cfg.gen_unit_cfg_header_file(cnf_header)
        LOG.info(
            "Finished generating compiler switches header file (in %4.2f s)",
            time.time() - start_time,
        )
        if build_cfg.has_yaml_interface:
            signal_if = YamlSignalInterfaces(build_cfg, unit_cfg, feature_cfg, udt)
        else:
            signal_if = CsvSignalInterfaces(build_cfg, unit_cfg)
            check_interfaces(build_cfg, signal_if)
        if interface:
            interface_report(build_cfg, unit_cfg, signal_if)

        LOG.info("******************************************************")
        LOG.info("Start generating VcExtVar and VcDebug")
        start_time = time.time()
        asil_level_db = build_defs.CVC_ASIL_LEVEL_MAP[code_generation_config["generalAsilLevelDebug"]]
        asil_level_dep = build_defs.ASIL_LEVEL_MAP[code_generation_config["generalAsilLevelDependability"]]
        nrm_dict, dep_dict, sec_dict, dbg_dict = signal_if.get_external_io()

        generate_external_var(build_cfg, unit_cfg, udt, asil_level_dep, nrm_dict, dep_dict, sec_dict)

        restructured_dbg_dict = {}
        dbg_instances = {
            ExtDbg(dbg_dict, build_cfg, unit_cfg): ("VcDebug", "VcDebugOutput"),
            ExtDbg(dep_dict, build_cfg, unit_cfg, asil_level_db): ("VcDebugSafe", "VcDebugOutputSafe")
        }
        for instance, dir_names in dbg_instances.items():
            restructured_dbg_dict.update({
                dir_names[0]: instance.dbg_dict["inputs"],
                dir_names[1]: instance.dbg_dict["outputs"]
            })
            instance.gen_dbg_files(
                pjoin(build_cfg.get_src_code_dst_dir(), dir_names[0]),
                pjoin(build_cfg.get_src_code_dst_dir(), dir_names[1])
            )
        LOG.info("Finished generating VcExtVar and VcDebug (in %4.2f s)", time.time() - start_time)

        if not code_generation_config["generateDummyVar"]:
            LOG.info("******************************************************")
            LOG.info("Skip generating VcDummy file")
        else:
            generate_dummy_var(build_cfg, unit_cfg, signal_if, udt)

        custom_dummy_spm = build_cfg.get_use_custom_dummy_spm()
        if custom_dummy_spm is not None:
            LOG.info("******************************************************")
            if os.path.isfile(custom_dummy_spm):
                LOG.info("Copying custom dummy spm file (%s)", custom_dummy_spm)
                shutil.copy2(custom_dummy_spm, build_cfg.get_src_code_dst_dir())
            else:
                LOG.warning(
                    "Cannot find desired custom dummy spm file: %s", custom_dummy_spm
                )
                generate_dummy_spm(build_cfg, unit_cfg, feature_cfg, signal_if, udt)
        else:
            generate_dummy_spm(build_cfg, unit_cfg, feature_cfg, signal_if, udt)

        custom_sources = build_cfg.get_use_custom_sources()
        if custom_sources is not None:
            LOG.info("******************************************************")
            for custom_src in custom_sources:
                if os.path.isfile(custom_src):
                    LOG.info("Copying custom sourcefile (%s)", custom_src)
                    shutil.copy2(custom_src, build_cfg.get_src_code_dst_dir())
                else:
                    LOG.warning("Cannot find desired custom sourcefile: %s", custom_src)

        LOG.info("******************************************************")
        LOG.info("Start generating the scheduling functions")
        start_time = time.time()
        gen_schd = SchedFuncs(build_cfg, unit_cfg, restructured_dbg_dict)
        gen_schd.generate_sched_c_fncs(generate_rte_checkpoint_calls)
        LOG.info("Finished generating the scheduling functions (in %4.2f s)", time.time() - start_time)

        LOG.info("******************************************************")
        LOG.info("Start generating the ts header file")
        start_time = time.time()
        gen_schd.generate_ts_defines(pjoin(src_dst_dir, build_cfg.get_ts_header_name()))
        LOG.info("Finished generating ts header file (in %4.2f s)", time.time() - start_time)

        # Generate AllSystemInfo.mat for DocGen compatibility
        if generate_system_info:
            from powertrain_build.gen_allsysteminfo import GenAllSystemInfo

            gen_all_system_info = GenAllSystemInfo(signal_if, unit_cfg)
            gen_all_system_info.build()

        # Check if errors
        if not no_abort:
            if problem_logger.errors():
                nbr_err = problem_logger.get_nbr_problems()
                problem_logger.info(
                    "Aborting build due to errors (# critical:%s, # warnings:%s"
                    " after %4.2f s.",
                    nbr_err["critical"],
                    nbr_err["warning"],
                    time.time() - tot_start_time,
                )
                return 1

        # Copy files to output folder
        copy_unit_src_to_src_out(build_cfg)
        copy_common_src_to_src_out(build_cfg)
        copy_unit_cfgs_to_output(build_cfg)
        copy_files_to_include(build_cfg)
        if code_generation_config["generateInterfaceHeaders"]:
            memory_section = MemorySection(build_cfg, unit_cfg)
            memory_section.generate_required_header_files()

        # Propagate tag name for release builds, TAG_NAME must be set in environment
        tag_name = os.environ.get("TAG_NAME", "")
        if tag_name and code_generation_config["propagateTagName"]:
            propagate_tag_name(build_cfg, tag_name, problem_logger)

        # Copy header files (subversion is using an external that points to
        # the correct set of pragma section header_files

        # Make A2L-file
        if generate_custom_conversion_table is not None:
            ctable_json = Path(generate_custom_conversion_table).resolve()
            ctable_a2l = Path(build_cfg.get_src_code_dst_dir(), "custom_tabs.a2l")
            create_conversion_table(ctable_json, ctable_a2l)
        merged_a2l = merge_a2l_files(build_cfg, unit_cfg, complete_a2l, silver_a2l)
        a2l_file_path = Path(build_cfg.get_src_code_dst_dir(), build_cfg.get_a2l_name())
        replace_tab_verb(a2l_file_path)

        # Generate interface files
        if code_generation_config["generateYamlInterfaceFile"]:
            zc_core = ZCCore(build_cfg, unit_cfg)
            zc_dids = ZCDIDs(build_cfg, unit_cfg)
            LOG.info("******************************************************")
            LOG.info("Start generating NVMDefinitions")
            start_time = time.time()
            project_nvm_defintions = unit_cfg.get_per_cfg_unit_cfg().get("nvm", {})
            zc_nvm = ZCNVMDef(build_cfg, unit_cfg, project_nvm_defintions)
            LOG.info(
                "Finished generating NVMDefinitions (in %4.2f s)", time.time() - start_time
            )
            LOG.info("******************************************************")
            axis_data = merged_a2l.get_characteristic_axis_data()
            composition_yaml = CompositionYaml(
                build_cfg,
                signal_if,
                unit_cfg,
                zc_core,
                zc_dids,
                zc_nvm,
                axis_data,
                udt.get_enumerations()
            )
            composition_yaml.generate_yaml()
            LOG.info("******************************************************")
            LOG.info("Generating Core header")
            zc_core.generate_dtc_files()
            LOG.info("******************************************************")
            LOG.info("Generating DID files")
            zc_dids.generate_did_files()
            LOG.info("******************************************************")
            LOG.info("Generating NVM definitions")
            zc_nvm.generate_nvm_rte_files()

            if rte_dummy:
                LOG.info("******************************************************")
                LOG.info("Generating RTE dummy files")
                zc_rte = RteDummy(build_cfg, zc_core, zc_nvm, composition_yaml.cal_class_info["tl"])
                zc_rte.generate_rte_dummy()

            if code_generation_config["generateCalibrationInterfaceFiles"]:
                LOG.info("******************************************************")
                if code_generation_config["useCalibrationRteMacroExpansion"]:
                    LOG.warning(
                        "Skip generating calibration interface files as useCalibrationRteMacroExpansion is set to true"
                    )
                else:
                    LOG.info("Generating calibration interface files")
                    zc_calibration = ZoneControllerCalibration(
                        build_cfg, composition_yaml.cal_class_info["tl"]
                    )
                    zc_calibration.generate_calibration_interface_files()
        elif build_cfg.get_ecu_info()[0] == "HI":
            generate_nvm_def(build_cfg, unit_cfg, no_nvm_a2l)
            LOG.info("******************************************************")
            LOG.info("Generating Core header")
            hi_core = HICore(build_cfg, unit_cfg)
            hi_core.generate_dtc_files()
            LOG.info("******************************************************")
            LOG.info("Generating DID files")
            dids = HIDIDs(build_cfg, unit_cfg)
            dids.generate_did_files()
        else:
            generate_nvm_def(build_cfg, unit_cfg, no_nvm_a2l)
            generate_did_files(build_cfg, unit_cfg)
            # generate core dummy files if requested
            if core_dummy:
                core_dummy_fname = os.path.basename(build_cfg.get_core_dummy_name())
                if CodeGenerators.embedded_coder in unit_cfg.code_generators:
                    LOG.info("******************************************************")
                    LOG.info("Skip generating %s for EC projects", core_dummy_fname)
                elif not code_generation_config["generateCoreDummy"]:
                    LOG.info("******************************************************")
                    LOG.info("Skip generating %s since generateCoreDummy is set to False", core_dummy_fname)
                else:
                    core = Core(build_cfg, unit_cfg)
                    generate_core_dummy(build_cfg, core, unit_cfg)

        if problem_logger.errors():
            problem_logger.info(
                "Critical errors were detected, aborting" " after %4.2f s.",
                time.time() - tot_start_time,
            )
            return 1

        LOG.info("Finished build in %4.2f s", time.time() - tot_start_time)
        return 0

    finally:
        logging.shutdown()



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module containing classes for VCC - Supplier signal interface."""

from powertrain_build import build_defs
from powertrain_build.types import byte_size_string, get_bitmask
from powertrain_build.a2l import A2l
from powertrain_build.problem_logger import ProblemLogger


class ExtVarBase(ProblemLogger):
    """Generate a2l- and c-files.

    These which declares all variables in the interface that the
    supplier platform writes to.

    This is needed due to legacy handling of the interfaces between units.
    Note that all variables sent from the VCC SPM to the platform should be declared in
    the function that produces the signal!
    """

    INPORT_INDEX = 0
    OUTPORT_INDEX = 1

    __data_type_size = {
        "Float32": "4",
        "UInt32": "4",
        "Int32": "4",
        "UInt16": "2",
        "Int16": "2",
        "UInt8": "1",
        "Int8": "1",
        "Bool": "1",
    }

    def __init__(self, variable_dict, prj_cfg, unit_cfg, user_defined_types, integrity_level=build_defs.ASIL_QM):
        """Constructor.

        Args:
            variable_dict (dict): dictionary with signal information.
            prj_cfg (BuildProjConfig): Build project class holding where files should be stored.
            user_defined_types (UserDefinedTypes): Class holding user defined data types.
            integrity_level (str): integrity level of the unit from 'A' to 'D' or 'QM'.
        """
        super().__init__()
        self.set_integrity_level(integrity_level)
        self._var_dict = variable_dict
        self._ext_vars = {}
        self._prj_cfg = prj_cfg
        self._unit_cfg = unit_cfg
        self._enumerations = user_defined_types.get_enumerations()
        self._common_header_files = user_defined_types.common_header_files

    def set_integrity_level(self, integrity_level):
        """Set integrity level of code generation.

        Args:
            integrity_level (str): integrity level of the unit from 'A' to 'D' or 'QM'
        """
        self._disp_start = integrity_level["CVC"]["DISP"]["START"]
        self._disp_end = integrity_level["CVC"]["DISP"]["END"]
        self._decl_start = integrity_level["PREDECL"]["DISP"]["START"]
        self._decl_end = integrity_level["PREDECL"]["DISP"]["END"]

    def _get_byte_size_string(self, data_type):
        """Get byte size of a data type as string.
        Enumeration byte sizes are derived from the underlying data type.

        Args:
            data_type (str): Data type.
        Returns:
            byte_size_string(powertrain_build.types.byte_size_string): Return result of
                                                                       powertrain_build.types.byte_size_string.
        """
        if data_type in self._enumerations:
            return byte_size_string(self._enumerations[data_type]["underlying_data_type"])
        return byte_size_string(data_type)

    def _get_bitmask(self, data_type):
        """Get bitmask of a data type.
        Enumeration bitmasks are derived from the underlying data type.

        Args:
            data_type (str): Data type.
        Returns:
            get_bitmask(powertrain_build.types.get_bitmask): Return result of powertrain_build.types.get_bitmask.
        """
        if data_type in self._enumerations:
            return get_bitmask(self._enumerations[data_type]["underlying_data_type"])
        return get_bitmask(data_type)

    def _restruct_input_data(self):
        """Restructure all the input variables per data-type.

        This will be used for declaring the variables and generating the
        A2L-file
        """
        external_inports = {}
        external_outports = {}
        for external_port_type in self.EXTERNAL_INPORT_TYPES:
            if external_port_type in self._var_dict:
                for var, data in self._var_dict[external_port_type].items():
                    data_type_size = self._get_byte_size_string(data[self.TYPE_NAME])
                    external_inports.setdefault(data_type_size, {})[var] = data
        for external_port_type in self.EXTERNAL_OUTPORT_TYPES:
            if external_port_type in self._var_dict:
                for var, data in self._var_dict[external_port_type].items():
                    data_type_size = self._get_byte_size_string(data[self.TYPE_NAME])
                    external_outports.setdefault(data_type_size, {})[var] = data
        self._ext_vars = external_inports, external_outports

    def _a2l_dict(self):
        """Return a dict defining all parameters for a2l-generation."""
        res = {"vars": {}, "function": "VcExtVar"}
        for inp in self.EXTERNAL_INPORT_TYPES:
            if inp in self._var_dict:
                for var, data in self._var_dict[inp].items():
                    if data[self.TYPE_NAME] in self._enumerations:
                        data_type = self._enumerations[data[self.TYPE_NAME]]["underlying_data_type"]
                    else:
                        data_type = data[self.TYPE_NAME]

                    resv = res["vars"]
                    resv.setdefault(var, {})["a2l_data"] = self.get_a2l_format(data)
                    resv[var]["array"] = []
                    resv[var]["function"] = ["VcEc"]
                    resv[var]["var"] = {"cvc_type": "CVC_DISP", "type": data_type, "var": var}
        return res

    def _get_array_declaration(self, var, width):
        """Get array declaration string from width information.

        Args:
            var (str): Variable name.
            width (int or list): Width information.
        Returns:
            array (str): Array declaration string.
        """
        array = ""
        widths = [width] if not isinstance(width, list) else width
        if len(widths) != 1 or widths[0] != 1:
            for w in widths:
                if w > 1:
                    if not isinstance(w, int):
                        self.critical(f'{var} widths must be integers. Got "{type(w)}"')
                    array += f'[{w}]'
                elif w < 0:
                    self.critical(f'{var} widths can not be negative. Got "{w}"')
        return array

    def _generate_c_file(self, path):
        """Generate the c-file defining all the supplier input signals."""
        header = path.with_suffix(".h").name
        var_set = set()
        with path.open("w") as fh_c:
            fh_c.write(f'#include "{header}"\n')
            fh_c.write(f'#include "{self._disp_start}"\n\n')
            for data_type_s, ext_vars in self._ext_vars[self.INPORT_INDEX].items():
                fh_c.write(f"/* Variables of size {data_type_s} bytes */\n\n")
                for var in sorted(ext_vars.keys()):
                    data = ext_vars[var]
                    if var not in var_set:
                        array = self._get_array_declaration(var, data.get("width", 1))
                        if array:
                            init_value = data["init"] if data["init"] != 0 else "{0}"
                        else:
                            init_value = data["init"]
                        fh_c.write(f"CVC_DISP {data[self.TYPE_NAME]} {var}{array} = {init_value};\n")
                        var_set.add(var)
                fh_c.write("\n")
            fh_c.write(f'\n#include "{self._disp_end}"\n')
            self.info("Generated %s", path.name)

    def _generate_h_file(self, path):
        """Generate header file externally declaring interface signals."""
        filename = path.stem
        guard = f"{filename.upper()}_H"
        var_set = set()
        with path.open("w") as fh_c:
            fh_c.write(f"#ifndef {guard}\n")
            fh_c.write(f"#define {guard}\n")
            fh_c.write("#define CVC_DISP\n")
            fh_c.write(self._unit_cfg.base_types_headers)

            for common_header_file in self._common_header_files:
                fh_c.write(f'#include "{common_header_file}"\n')
            fh_c.write("\n")

            fh_c.write(f'#include "{self._decl_start}"\n')
            fh_c.write("/* VCC Inports */\n")
            for data_type_s, ext_vars in self._ext_vars[self.INPORT_INDEX].items():
                fh_c.write(f"/* Variables of size {data_type_s} bytes */\n\n")
                for var in sorted(ext_vars.keys()):
                    if var not in var_set:
                        data = ext_vars[var]
                        array = self._get_array_declaration(var, data.get("width", 1))
                        fh_c.write(f"extern CVC_DISP {data[self.TYPE_NAME]} {var}{array};\n")
                        var_set.add(var)
                fh_c.write("\n")

            fh_c.write("/* VCC Outports */\n")
            for data_type_s, ext_vars in self._ext_vars[self.OUTPORT_INDEX].items():
                fh_c.write(f"/* Variables of size {data_type_s} bytes */\n\n")
                for var in sorted(ext_vars.keys()):
                    if var not in var_set:
                        data = ext_vars[var]
                        array = self._get_array_declaration(var, data.get("width", 1))
                        fh_c.write(f"extern CVC_DISP {data[self.TYPE_NAME]} {var}{array};\n")
                        var_set.add(var)
                fh_c.write("\n")
            fh_c.write(f'#include "{self._decl_end}"\n')
            fh_c.write("#endif\n")
            self.info("Generated %s", path.name)

    def generate_files(self, path):
        """Generate the c- and a2l-file for defining all the supplier input variables."""
        self._restruct_input_data()
        if not self._ext_vars[0] and not self._ext_vars[1]:
            self.info(f"Skipping {path.name} as there were no corresponding vars.")
            return
        self._generate_c_file(path.with_suffix(".c"))
        self._generate_h_file(path.with_suffix(".h"))
        a2l_dict = self._a2l_dict()
        a2l = A2l(a2l_dict, self._prj_cfg)
        a2l.gen_a2l(path.with_suffix(".a2l"))


class ExtVarCsv(ExtVarBase):
    """Handles variable dicts from CSV files.

    Variable dict shall have the following format and is generated by the
    :doc:`CsvSignalInterfaces <signal_interfaces>` class::

        {
            'CAN-Input': {
                'signal1': {
                    'IOType': 'd',
                    'description': 'Some description',
                    'init': 0,
                    'max': 1,
                    'min': 0,
                    'type': 'UInt8',
                    'unit': '-'
                },
                'signal2': {
                    ...
                }
            },
            'CAN-Output': {
                'signal3': {
                    ...
                }
            },
            'xxx-Input': ...,
            'xxx-Output': ...
        }
    """

    EXTERNAL_INPORT_TYPES = ["EMS-Input", "CAN-Input", "Private CAN-Input", "LIN-Input"]
    EXTERNAL_OUTPORT_TYPES = ["EMS-Output", "CAN-Output", "Private CAN-Output", "LIN-Output"]
    TYPE_NAME = "type"

    def get_a2l_format(self, data):
        """Get a2l format.

        Args:
            data (dict): Data dictionary.
        Returns:
            dict: A2l format dictionary.
        """
        return {
            "bitmask": self._get_bitmask(data[self.TYPE_NAME]),
            "description": data["description"],
            "lsb": "2^0",
            "max": data["max"],
            "min": data["min"],
            "offset": "0",
            "unit": data["unit"],
            "x_axis": None,
            "y_axis": None,
        }


class ExtVarYaml(ExtVarBase):
    """Handles variable dicts from Yaml files.

    Variable dict shall have the following format and is generated by the
    :doc:`YamlSignalInterfaces <signal_interfaces>` class::

        {
            'input': {
                'sVcIhfa_D_WhlMotSysFrntLimnIndcn': {},
                'sVcIhfa_D_WhlMotSysFrntModSts': {},
                'sVcIhfa_I_WhlMotSysFrntIdc': {},
                'sVcIhfa_U_UDcDcActHiSide1': {},
                'sVcIhfa_U_WhlMotSysFrntUdc': {}
            },
            'output': {},
            'status': {},
        }
    """

    EXTERNAL_INPORT_TYPES = ["input", "status"]
    EXTERNAL_OUTPORT_TYPES = ["output"]
    TYPE_NAME = "variable_type"

    def get_a2l_format(self, data):
        """Get a2l format.

        Args:
            data (dict): Data dictionary.
        Returns:
            dict: A2l format dictionary.
        """
        return {
            "bitmask": self._get_bitmask(data[self.TYPE_NAME]),
            "description": data["description"],
            "lsb": "2^0",
            "max": data["range"]["max"],
            "min": data["range"]["min"],
            "offset": "0",
            "unit": data["unit"],
            "x_axis": None,
            "y_axis": None,
        }
